import research_effects;
import util.design_export;

from traits import getTraitID;

//DOF - Allow detection of Traits in Research
tidy final class CheckHasTrait : TechnologyHook {
	Document doc("Only apply the inner hook if the empire also has another trait. Cannot be used for continuous effects, only ones applied once at game start.");
	Argument trait(AT_Trait, doc="Trait to check for.");
	Argument function(AT_Hook, "research_effects::ITechnologyHook");
	//Argument hookID("Hook", AT_Hook, "trait_effects::TraitEffect");
	//Argument elseID("Else", AT_Hook, "", "trait_effects::TraitEffect");

#section server
	ITechnologyHook@ hook;
	
	bool instantiate() override {
		@hook = cast<ITechnologyHook>(parseHook(function.str, "research_effects::", required=false));
		if(hook is null) {
			error("NotSecondary(): could not find inner hook: "+escape(function.str));
			return false;
		}
		return TechnologyHook::instantiate();
	}

	void unlock(TechnologyNode@ node, Empire& emp) const {
		if(emp.hasTrait(trait.integer))  {
			//print("TraitCheck = TRUE");
			hook.unlock(node, emp);
		}
		//else  print("TraitCheck = FALSE");
	}
#section all
};

//DOF - Design import from Research
class LoadResearchedDesigns : TechnologyHook {
	Document doc("Load default designs from a particular directory into the empire.");
	Argument directory(AT_Custom, doc="Relative path to the directory to add designs from.");
	Argument limit_shipset(AT_Boolean, "True", doc="Whether to only load designs that have a saved hull that matches the current shipset.");
	Argument retry_without_limit(AT_Boolean, "True", doc="Whether to override the shipset limit and load all designs if no designs were found.");

#section server
	DesignSet designs;

	bool instantiate() override {
		designs.readDirectory("data/designs/"+directory.str);
		designs.limitShipset = limit_shipset.boolean;
		designs.softLimitRetry = retry_without_limit.boolean;
		return TechnologyHook::instantiate();
	}

	void unlock(TechnologyNode@ node, Empire& emp) const override {
		designs.createFor(emp);
	}
#section all
};


//Better Version
class LoadResearchedDesignsAuto : TechnologyHook {
	Document doc("Load Support, Flagship & Mothership designs for the appropriate lifestyle and FTL type from the /advanced subfolder of each option, following the standard structure.");
	Argument limit_shipset(AT_Boolean, "True", doc="Whether to only load designs that have a saved hull that matches the current shipset.");
	Argument retry_without_limit(AT_Boolean, "True", doc="Whether to override the shipset limit and load all designs if no designs were found.");

#section server

	void unlock(TechnologyNode@ node, Empire& emp) const override {
		DesignSet designs;
		DesignSet designsSupport;
		DesignSet designsMothership;
		bool bMotherships = false;
		string target = "/advanced";
		string path = "data/designs";
		string ftl = "";
		//FTL
		if(emp.hasTrait(getTraitID("Hyperdrive")))  ftl = "/hyperdrive";
		else if(emp.hasTrait(getTraitID("Jumpdrive")))  ftl = "/jumpdrive";
		else  ftl = "/not_hyperdrive";
		//Lifestyle
		//Stock Races
		if(emp.hasTrait(getTraitID("Ancient")))  path = path+"/ancient";
		else if(emp.hasTrait(getTraitID("Devout")))  path = path+"/devout";
		else if(emp.hasTrait(getTraitID("Starchildren")))  {
			bMotherships = true;
		}
		else if(emp.hasTrait(getTraitID("Verdant")))  path = path+"/verdant";
		//DOF Specific Races
		else if(emp.hasTrait(getTraitID("Ancients")))  path = path+"/Ancients";
		else if(emp.hasTrait(getTraitID("AncientsSojourners")))  {
			path = path+"/Ancients";
			bMotherships = true;
		}
		else if(emp.hasTrait(getTraitID("AncientsCannonical")))  path = path+"/AncientsCannonical";
		else if(emp.hasTrait(getTraitID("Asgard")))  path = path+"/Asgard";
		else if(emp.hasTrait(getTraitID("BSGColonial")))  path = path+"/BSGColonial";
		else if(emp.hasTrait(getTraitID("BSGCylon")))  path = path+"/BSGCylon";
		else if(emp.hasTrait(getTraitID("SWCIS")))  path = path+"/CIS";
		else if(emp.hasTrait(getTraitID("SWEmpire")))  path = path+"/Empire";
		else if(emp.hasTrait(getTraitID("Goauld")))  path = path+"/Goauld";
		else if(emp.hasTrait(getTraitID("SWRepublic")))  path = path+"/Republic";
		else if(emp.hasTrait(getTraitID("STFederation")))  path = path+"/STFederation";
		else if(emp.hasTrait(getTraitID("STKlingon")))  path = path+"/STKlingon";
		else if(emp.hasTrait(getTraitID("STBorg")))  path = path+"/STBorg";
		else if(emp.hasTrait(getTraitID("Tauri")))  path = path+"/Tauri";
		else if(emp.hasTrait(getTraitID("Wraith")))  {
			path = path+"/Wraith";
			bMotherships = true;
		}
		//print("Path = "+path+ftl+target);		
		//Flagships
		designs.readDirectory(path+ftl+target);
		designs.limitShipset = limit_shipset.boolean;
		designs.softLimitRetry = retry_without_limit.boolean;
		designs.createFor(emp);
		//Supports
		designsSupport.readDirectory(path+"/default"+target);
		designsSupport.limitShipset = limit_shipset.boolean;
		designsSupport.softLimitRetry = retry_without_limit.boolean;
		designsSupport.createFor(emp);
		//Motherships
		if (bMotherships)  {
			if(ftl == "/not_hyperdrive")  {
				if(emp.hasTrait(getTraitID("Slipstream")))  ftl = "/slipstream";
				else  ftl = "/other";
			}
			if(emp.hasTrait(getTraitID("Starchildren")))  path = "data/designs/starchildren";
			else  path = path+"/mothership";
			designsMothership.readDirectory(path+ftl+target);
			designsMothership.limitShipset = limit_shipset.boolean;
			designsMothership.softLimitRetry = retry_without_limit.boolean;
			designsMothership.createFor(emp);
		}
	}
#section all
};

/*
//DOF - Design import from Research
class LoadResearchedDesignsAuto : TechnologyHook {
	Document doc("Load default designs from a particular directory into the empire.");
	Argument limit_shipset(AT_Boolean, "True", doc="Whether to only load designs that have a saved hull that matches the current shipset.");
	Argument retry_without_limit(AT_Boolean, "True", doc="Whether to override the shipset limit and load all designs if no designs were found.");

#section server
	DesignSet designs;
	DesignSet designsSupport;
	DesignSet designsMothership;
	
	bool instantiate() override {
		Empire@ emp = defaultEmpire;
		bool bMotherships = false;
		string path = "data/designs";
		string ftl = "";
		//FTL
		if(emp.hasTrait(getTraitID("Hyperdrive")))  ftl = "/hyperdrive";
		else if(emp.hasTrait(getTraitID("Jumpdrive")))  ftl = "/jumpdrive";
		else  ftl = "/not_hyperdrive";
		//Lifestyle
		if(emp.hasTrait(getTraitID("Ancient")))  path = path+"/ancient";
		else if(emp.hasTrait(getTraitID("Devout")))  path = path+"/devout";
		else if(emp.hasTrait(getTraitID("Starchildren")))  {
			bMotherships = true;
		}
		else if(emp.hasTrait(getTraitID("Verdant")))  path = path+"/verdant";
		else if(emp.hasTrait(getTraitID("Ancients")))  path = path+"/Ancients";
		else if(emp.hasTrait(getTraitID("AncientsSojourners")))  {
			path = path+"/Ancients";
			bMotherships = true;
		}
		else if(emp.hasTrait(getTraitID("AncientsCannonical")))  path = path+"/AncientsCannonical";
		else if(emp.hasTrait(getTraitID("Asgard")))  path = path+"/Asgard";
		else if(emp.hasTrait(getTraitID("BSGColonial")))  path = path+"/BSGColonial";
		else if(emp.hasTrait(getTraitID("BSGCylon")))  path = path+"/BSGCylon";
		else if(emp.hasTrait(getTraitID("Empire")))  path = path+"/Empire";
		else if(emp.hasTrait(getTraitID("Goauld")))  path = path+"/Goauld";
		else if(emp.hasTrait(getTraitID("Republic")))  path = path+"/Republic";
		else if(emp.hasTrait(getTraitID("STFederation")))  path = path+"/STFederation";
		else if(emp.hasTrait(getTraitID("Tauri")))  path = path+"/Tauri";
		else if(emp.hasTrait(getTraitID("Wraith")))  {
			path = path+"/Wraith";
			bMotherships = true;
		}
		print("Path = "+path+ftl);
		//Flagships
		designs.readDirectory(path+ftl+"/advanced");
		designs.limitShipset = limit_shipset.boolean;
		designs.softLimitRetry = retry_without_limit.boolean;
		//Supports
		designsSupport.readDirectory(path+"/default/advanced");
		designsSupport.limitShipset = limit_shipset.boolean;
		designsSupport.softLimitRetry = retry_without_limit.boolean;
		//Motherships
		if (bMotherships)  {
			if(ftl == "/not_hyperdrive")  {
				if(emp.hasTrait(getTraitID("Slipstream")))  ftl = "/slipstream";
				else  ftl = "/other";
			}
			if(emp.hasTrait(getTraitID("Starchildren")))  path = "data/designs/starchildren";
			else  path = path+"/mothership";
			designsMothership.readDirectory(path+ftl+"/advanced");
			designsMothership.limitShipset = limit_shipset.boolean;
			designsMothership.softLimitRetry = retry_without_limit.boolean;
		}
		return TechnologyHook::instantiate();
	}
	void unlock(TechnologyNode@ node, Empire& emp) const override {
		designs.createFor(emp);
		designsSupport.createFor(emp);
		designsMothership.createFor(emp);
	}
#section all
}

	
class LoadResearchedDesignsAuto : TechnologyHook {
	Document doc("Load Support, Flagship & Mothership designs for the appropriate lifestyle and FTL type from the /advanced subfolder of each option, following the standard structure.  Bloated/Costly and should only be used on Nodes shared between multiple lifestyles.");
	Argument limit_shipset(AT_Boolean, "True", doc="Whether to only load designs that have a saved hull that matches the current shipset.");
	Argument retry_without_limit(AT_Boolean, "True", doc="Whether to override the shipset limit and load all designs if no designs were found.");

#section server
	DesignSet VerdantSupport;
	DesignSet VerdantFlagshipHyperdrive;
	DesignSet VerdantFlagshipJumpdrive;
	DesignSet VerdantFlagshipNotHyperdrive;
	DesignSet AncientsSupport;
	DesignSet AncientsFlagshipHyperdrive;
	DesignSet AncientsFlagshipJumpdrive;
	DesignSet AncientsFlagshipNotHyperdrive;
	DesignSet AncientsMothershipHyperdrive;
	DesignSet AncientsMothershipJumpdrive;
	DesignSet AncientsMothershipSlipstream;
	DesignSet AncientsMothershipNotHyperdrive;
	DesignSet AncientsCannonicalSupport;
	DesignSet AncientsCannonicalFlagshipHyperdrive;
	DesignSet AncientsCannonicalFlagshipJumpdrive;
	DesignSet AncientsCannonicalFlagshipNotHyperdrive;
	DesignSet AsgardSupport;
	DesignSet AsgardFlagshipHyperdrive;
	DesignSet AsgardFlagshipJumpdrive;
	DesignSet AsgardFlagshipNotHyperdrive;
	DesignSet BSGColonialSupport;
	DesignSet BSGColonialFlagshipHyperdrive;
	DesignSet BSGColonialFlagshipJumpdrive;
	DesignSet BSGColonialFlagshipNotHyperdrive;
	DesignSet BSGCylonSupport;
	DesignSet BSGCylonFlagshipHyperdrive;
	DesignSet BSGCylonFlagshipJumpdrive;
	DesignSet BSGCylonFlagshipNotHyperdrive;
	DesignSet EmpireSupport;
	DesignSet EmpireFlagshipHyperdrive;
	DesignSet EmpireFlagshipJumpdrive;
	DesignSet EmpireFlagshipNotHyperdrive;
	DesignSet GoauldSupport;
	DesignSet GoauldFlagshipHyperdrive;
	DesignSet GoauldFlagshipJumpdrive;
	DesignSet GoauldFlagshipNotHyperdrive;
	DesignSet RepublicSupport;
	DesignSet RepublicFlagshipHyperdrive;
	DesignSet RepublicFlagshipJumpdrive;
	DesignSet RepublicFlagshipNotHyperdrive;
	DesignSet STFederationSupport;
	DesignSet STFederationFlagshipHyperdrive;
	DesignSet STFederationFlagshipJumpdrive;
	DesignSet STFederationFlagshipNotHyperdrive;
	DesignSet TauriSupport;
	DesignSet TauriFlagshipHyperdrive;
	DesignSet TauriFlagshipJumpdrive;
	DesignSet TauriFlagshipNotHyperdrive;
	DesignSet WraithSupport;
	DesignSet WraithFlagshipHyperdrive;
	DesignSet WraithFlagshipJumpdrive;
	DesignSet WraithFlagshipNotHyperdrive;
	DesignSet WraithMothershipHyperdrive;
	DesignSet WraithMothershipJumpdrive;
	DesignSet WraithMothershipSlipstream;
	DesignSet WraithMothershipNotHyperdrive;
	
	bool instantiate() override {
		//Verdant
		VerdantSupport.readDirectory("data/designs/verdant/default/advanced");
		VerdantSupport.limitShipset = limit_shipset.boolean;
		VerdantSupport.softLimitRetry = retry_without_limit.boolean;
		VerdantFlagshipHyperdrive.readDirectory("data/designs/verdant/hyperdrive/advanced");
		VerdantFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		VerdantFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		VerdantFlagshipJumpdrive.readDirectory("data/designs/verdant/jumpdrive/advanced");
		VerdantFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		VerdantFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		VerdantFlagshipNotHyperdrive.readDirectory("data/designs/verdant/not_hyperdrive/advanced");
		VerdantFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		VerdantFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Ancients & Ancients Sojourners
		AncientsSupport.readDirectory("data/designs/Ancients/default/advanced");
		AncientsSupport.limitShipset = limit_shipset.boolean;
		AncientsSupport.softLimitRetry = retry_without_limit.boolean;
		AncientsFlagshipHyperdrive.readDirectory("data/designs/Ancients/hyperdrive/advanced");
		AncientsFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		AncientsFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		AncientsFlagshipJumpdrive.readDirectory("data/designs/Ancients/jumpdrive/advanced");
		AncientsFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		AncientsFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		AncientsFlagshipNotHyperdrive.readDirectory("data/designs/Ancients/not_hyperdrive/advanced");
		AncientsFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		AncientsFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		AncientsMothershipHyperdrive.readDirectory("data/designs/Ancients/mothership/hyperdrive/advanced");
		AncientsMothershipHyperdrive.limitShipset = limit_shipset.boolean;
		AncientsMothershipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		AncientsMothershipJumpdrive.readDirectory("data/designs/Ancients/mothership/jumpdrive/advanced");
		AncientsMothershipJumpdrive.limitShipset = limit_shipset.boolean;
		AncientsMothershipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		AncientsMothershipSlipstream.readDirectory("data/designs/Ancients/mothership/slipstream/advanced");
		AncientsMothershipSlipstream.limitShipset = limit_shipset.boolean;
		AncientsMothershipSlipstream.softLimitRetry = retry_without_limit.boolean;
		AncientsMothershipNotHyperdrive.readDirectory("data/designs/Ancients/mothership/not_hyperdrive/advanced");
		AncientsMothershipNotHyperdrive.limitShipset = limit_shipset.boolean;
		AncientsMothershipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Cannonical Ancients
		AncientsCannonicalSupport.readDirectory("data/designs/AncientsCannonical/default/advanced");
		AncientsCannonicalSupport.limitShipset = limit_shipset.boolean;
		AncientsCannonicalSupport.softLimitRetry = retry_without_limit.boolean;
		AncientsCannonicalFlagshipHyperdrive.readDirectory("data/designs/AncientsCannonical/hyperdrive/advanced");
		AncientsCannonicalFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		AncientsCannonicalFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		AncientsCannonicalFlagshipJumpdrive.readDirectory("data/designs/AncientsCannonical/jumpdrive/advanced");
		AncientsCannonicalFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		AncientsCannonicalFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		AncientsCannonicalFlagshipNotHyperdrive.readDirectory("data/designs/AncientsCannonical/not_hyperdrive/advanced");
		AncientsCannonicalFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		AncientsCannonicalFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Asgard
		AsgardSupport.readDirectory("data/designs/Asgard/default/advanced");
		AsgardSupport.limitShipset = limit_shipset.boolean;
		AsgardSupport.softLimitRetry = retry_without_limit.boolean;
		AsgardFlagshipHyperdrive.readDirectory("data/designs/Asgard/hyperdrive/advanced");
		AsgardFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		AsgardFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		AsgardFlagshipJumpdrive.readDirectory("data/designs/Asgard/jumpdrive/advanced");
		AsgardFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		AsgardFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		AsgardFlagshipNotHyperdrive.readDirectory("data/designs/Asgard/not_hyperdrive/advanced");
		AsgardFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		AsgardFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//BSGColonial
		BSGColonialSupport.readDirectory("data/designs/BSGColonial/default/advanced");
		BSGColonialSupport.limitShipset = limit_shipset.boolean;
		BSGColonialSupport.softLimitRetry = retry_without_limit.boolean;
		BSGColonialFlagshipHyperdrive.readDirectory("data/designs/BSGColonial/hyperdrive/advanced");
		BSGColonialFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		BSGColonialFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		BSGColonialFlagshipJumpdrive.readDirectory("data/designs/BSGColonial/jumpdrive/advanced");
		BSGColonialFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		BSGColonialFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		BSGColonialFlagshipNotHyperdrive.readDirectory("data/designs/BSGColonial/not_hyperdrive/advanced");
		BSGColonialFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		BSGColonialFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//BSGCylonAsgardSupport.readDirectory("data/designs/Asgard/default/advanced");
		BSGCylonSupport.limitShipset = limit_shipset.boolean;
		BSGCylonSupport.softLimitRetry = retry_without_limit.boolean;
		BSGCylonFlagshipHyperdrive.readDirectory("data/designs/BSGCylon/hyperdrive/advanced");
		BSGCylonFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		BSGCylonFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		BSGCylonFlagshipJumpdrive.readDirectory("data/designs/BSGCylon/jumpdrive/advanced");
		BSGCylonFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		BSGCylonFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		BSGCylonFlagshipNotHyperdrive.readDirectory("data/designs/BSGCylon/not_hyperdrive/advanced");
		BSGCylonFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		BSGCylonFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Empire
		EmpireSupport.readDirectory("data/designs/Empire/default/advanced");
		EmpireSupport.limitShipset = limit_shipset.boolean;
		EmpireSupport.softLimitRetry = retry_without_limit.boolean;
		EmpireFlagshipHyperdrive.readDirectory("data/designs/Empire/hyperdrive/advanced");
		EmpireFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		EmpireFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		EmpireFlagshipJumpdrive.readDirectory("data/designs/Empire/jumpdrive/advanced");
		EmpireFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		EmpireFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		EmpireFlagshipNotHyperdrive.readDirectory("data/designs/Empire/not_hyperdrive/advanced");
		EmpireFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		EmpireFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Goauld
		GoauldSupport.readDirectory("data/designs/Goauld/default/advanced");
		GoauldSupport.limitShipset = limit_shipset.boolean;
		GoauldSupport.softLimitRetry = retry_without_limit.boolean;
		GoauldFlagshipHyperdrive.readDirectory("data/designs/Goauld/hyperdrive/advanced");
		GoauldFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		GoauldFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		GoauldFlagshipJumpdrive.readDirectory("data/designs/Goauld/jumpdrive/advanced");
		GoauldFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		GoauldFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		GoauldFlagshipNotHyperdrive.readDirectory("data/designs/Goauld/not_hyperdrive/advanced");
		GoauldFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		GoauldFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Republic
		RepublicSupport.readDirectory("data/designs/Republic/default/advanced");
		RepublicSupport.limitShipset = limit_shipset.boolean;
		RepublicSupport.softLimitRetry = retry_without_limit.boolean;
		RepublicFlagshipHyperdrive.readDirectory("data/designs/Republic/hyperdrive/advanced");
		RepublicFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		RepublicFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		RepublicFlagshipJumpdrive.readDirectory("data/designs/Republic/jumpdrive/advanced");
		RepublicFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		RepublicFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		RepublicFlagshipNotHyperdrive.readDirectory("data/designs/Republic/not_hyperdrive/advanced");
		RepublicFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		RepublicFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//STFederation
		STFederationSupport.readDirectory("data/designs/STFederation/default/advanced");
		STFederationSupport.limitShipset = limit_shipset.boolean;
		STFederationSupport.softLimitRetry = retry_without_limit.boolean;
		STFederationFlagshipHyperdrive.readDirectory("data/designs/STFederation/hyperdrive/advanced");
		STFederationFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		STFederationFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		STFederationFlagshipJumpdrive.readDirectory("data/designs/STFederation/jumpdrive/advanced");
		STFederationFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		STFederationFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		STFederationFlagshipNotHyperdrive.readDirectory("data/designs/STFederation/not_hyperdrive/advanced");
		STFederationFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		STFederationFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Tauri
		TauriSupport.readDirectory("data/designs/Tauri/default/advanced");
		TauriSupport.limitShipset = limit_shipset.boolean;
		TauriSupport.softLimitRetry = retry_without_limit.boolean;
		TauriFlagshipHyperdrive.readDirectory("data/designs/Tauri/hyperdrive/advanced");
		TauriFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		TauriFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		TauriFlagshipJumpdrive.readDirectory("data/designs/Tauri/jumpdrive/advanced");
		TauriFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		TauriFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		TauriFlagshipNotHyperdrive.readDirectory("data/designs/Tauri/not_hyperdrive/advanced");
		TauriFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		TauriFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		//Wraith
		WraithSupport.readDirectory("data/designs/Wraith/default/advanced");
		WraithSupport.limitShipset = limit_shipset.boolean;
		WraithSupport.softLimitRetry = retry_without_limit.boolean;
		WraithFlagshipHyperdrive.readDirectory("data/designs/Wraith/hyperdrive/advanced");
		WraithFlagshipHyperdrive.limitShipset = limit_shipset.boolean;
		WraithFlagshipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		WraithFlagshipJumpdrive.readDirectory("data/designs/Wraith/jumpdrive/advanced");
		WraithFlagshipJumpdrive.limitShipset = limit_shipset.boolean;
		WraithFlagshipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		WraithFlagshipNotHyperdrive.readDirectory("data/designs/Wraith/not_hyperdrive/advanced");
		WraithFlagshipNotHyperdrive.limitShipset = limit_shipset.boolean;
		WraithFlagshipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		WraithMothershipHyperdrive.readDirectory("data/designs/Wraith/mothership/hyperdrive/advanced");
		WraithMothershipHyperdrive.limitShipset = limit_shipset.boolean;
		WraithMothershipHyperdrive.softLimitRetry = retry_without_limit.boolean;
		WraithMothershipJumpdrive.readDirectory("data/designs/Wraith/mothership/jumpdrive/advanced");
		WraithMothershipJumpdrive.limitShipset = limit_shipset.boolean;
		WraithMothershipJumpdrive.softLimitRetry = retry_without_limit.boolean;
		WraithMothershipSlipstream.readDirectory("data/designs/Wraith/mothership/slipstream/advanced");
		WraithMothershipSlipstream.limitShipset = limit_shipset.boolean;
		WraithMothershipSlipstream.softLimitRetry = retry_without_limit.boolean;
		WraithMothershipNotHyperdrive.readDirectory("data/designs/Wraith/mothership/not_hyperdrive/advanced");
		WraithMothershipNotHyperdrive.limitShipset = limit_shipset.boolean;
		WraithMothershipNotHyperdrive.softLimitRetry = retry_without_limit.boolean;
		
		return TechnologyHook::instantiate();
	}

	void unlock(TechnologyNode@ node, Empire& emp) const override {
		if(emp.hasTrait(getTraitID("Verdant")))  {
			VerdantSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  VerdantFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  VerdantFlagshipJumpdrive.createFor(emp);
			else  VerdantFlagshipNotHyperdrive.createFor(emp);		
		}
		else if(emp.hasTrait(getTraitID("Ancients")))  {
			AncientsSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  AncientsFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  AncientsFlagshipJumpdrive.createFor(emp);
			else  AncientsFlagshipNotHyperdrive.createFor(emp);		
		}
		else if(emp.hasTrait(getTraitID("AncientsSojourners")))  {
			AncientsSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  {
				AncientsFlagshipHyperdrive.createFor(emp);
				AncientsMothershipHyperdrive.createFor(emp);
			}
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  {
				AncientsFlagshipJumpdrive.createFor(emp);
				AncientsMothershipJumpdrive.createFor(emp);
			}
			else  {
				AncientsFlagshipNotHyperdrive.createFor(emp);
				if(emp.hasTrait(getTraitID("Slipstream")))  AncientsMothershipSlipstream.createFor(emp);
				else  AncientsMothershipNotHyperdrive.createFor(emp);
			}
		}
		else if(emp.hasTrait(getTraitID("Asgard")))  {
			AsgardSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  AsgardFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  AsgardFlagshipJumpdrive.createFor(emp);
			else  AsgardFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("BSGColonial")))  {
			BSGColonialSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  BSGColonialFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  BSGColonialFlagshipJumpdrive.createFor(emp);
			else  BSGColonialFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("BSGCylon")))  {
			BSGCylonSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  BSGCylonFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  BSGCylonFlagshipJumpdrive.createFor(emp);
			else  BSGCylonFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("Empire")))  {
			EmpireSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  EmpireFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  EmpireFlagshipJumpdrive.createFor(emp);
			else  EmpireFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("Goauld")))  {
			GoauldSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  GoauldFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  GoauldFlagshipJumpdrive.createFor(emp);
			else  GoauldFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("Republic")))  {
			RepublicSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  RepublicFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  RepublicFlagshipJumpdrive.createFor(emp);
			else  RepublicFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("STFederation")))  {
			STFederationSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  STFederationFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  STFederationFlagshipJumpdrive.createFor(emp);
			else  STFederationFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("Tauri")))  {
			TauriSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  TauriFlagshipHyperdrive.createFor(emp);
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  TauriFlagshipJumpdrive.createFor(emp);
			else  TauriFlagshipNotHyperdrive.createFor(emp);
		}
		else if(emp.hasTrait(getTraitID("Wraith")))  {
			WraithSupport.createFor(emp);
			if(emp.hasTrait(getTraitID("Hyperdrive")))  {
				WraithFlagshipHyperdrive.createFor(emp);
				WraithMothershipHyperdrive.createFor(emp);
			}
			else if(emp.hasTrait(getTraitID("Jumpdrive")))  {
				WraithFlagshipJumpdrive.createFor(emp);
				WraithMothershipJumpdrive.createFor(emp);
			}
			else  {
				WraithFlagshipNotHyperdrive.createFor(emp);
				if(emp.hasTrait(getTraitID("Slipstream")))  WraithMothershipSlipstream.createFor(emp);
				else  WraithMothershipNotHyperdrive.createFor(emp);
			}
		}
	}
#section all
};
*/

class ResearchModEmpireAttribute : TechnologyHook {
	Document doc("Modifies or adds the specified empire attribute in the mode specified.");
	Argument attribute(AT_EmpAttribute, doc="Which attribute to alter.");
	Argument mode(AT_AttributeMode, doc="How to modify the attribute.");
	Argument value(AT_Decimal, doc="Value to modify the attribute by.");
	
#section server
	void unlock(TechnologyNode@ node, Empire& emp) const {
		if(emp !is null && emp.valid)
			emp.modAttribute(uint(arguments[0].integer), arguments[1].integer, arguments[2].decimal);
	}
#section all
};
