import empire_ai.weasel.WeaselAI;
import empire_ai.weasel.race.Race;
import empire_ai.weasel.Designs;

import empire_ai.weasel.Military;
import empire_ai.weasel.Construction;
import empire_ai.weasel.Development;
import empire_ai.weasel.Budget;

//DOF
//from traits import getTraitID;
from attributes import getEmpAttribute;

class Asgard : Race, RaceDesigns {
	Designs@ designs;
	
	Fleets@ fleets;
	Military@ military;
	Development@ development;
	Construction@ construction;
	Budget@ budget;

	//Designs from Verdant AI
	array<const Design@> defaultDesigns;
	array<uint> defaultGoals;
	
	//DOF - Design inventory tracking
	array<const Design@> InventoriedDesigns;
	uint DesignLevel = 0;
	
	//array<FleetAI@> valhallaStations;
	array<Object@> valhallaStations;
	const Design@ vsL1, vsL2, vsL3, vsL4, vsL5, vsCurrent;
	int vsScalingFactor = 4;
	uint vsScalingIntervalCount = 0;
	BuildStation@ vsBuild;
	AllocateConstruction@ vsRetrofit;
	Factory@ factory;

	void create() override {
		//Adjust size per money for better behavior
		ai.behavior.shipSizePerMoney = 1.0 / 6.00;
		ai.behavior.shipSizePerLabor = 1.0 / 0.65;
		
		@designs = cast<Designs>(ai.designs);
		
		@fleets = cast<Fleets>(ai.fleets);
		@military = cast<Military>(ai.military);
		@development = cast<Development>(ai.development);
		@construction = cast<Construction>(ai.construction);
		@budget = cast<Budget>(ai.budget);

	}

	void start() override {
		ReadLock lock(ai.empire.designMutex);
		for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
			const Design@ dsg = ai.empire.getDesign(i);
			
			//DOF - Add design to inventoried list
			InventoriedDesigns.insertLast(dsg);
			
			if(dsg.newer !is null)
				continue;
			if(dsg.updated !is null)
				continue;

			uint goal = designs.classify(dsg, DP_Unknown);
			if(goal == DP_Unknown)
				continue;

			defaultDesigns.insertLast(dsg);
			defaultGoals.insertLast(goal);
		}
		
		@vsL1 = ai.empire.getDesign("ASG Valhalla Station");
		@vsL2 = ai.empire.getDesign("ASG Valhalla Station Mk II");
		@vsL3 = ai.empire.getDesign("ASG Valhalla Station Mk III");
		@vsL4 = ai.empire.getDesign("ASG Valhalla Station Mk IV");
		@vsL5 = ai.empire.getDesign("ASG Valhalla Station Mk V");
	}

	void save(SaveFile& file) override {
		uint cnt = defaultDesigns.length;
		file << cnt;
		for(uint i = 0; i < cnt; ++i) {
			file << defaultDesigns[i];
			file << defaultGoals[i];
		}
		
		//DOF
		file << DesignLevel;
		uint idCnt = InventoriedDesigns.length;
		//print("idCnt = "+idCnt);
		file << idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file << InventoriedDesigns[i];
		}
		
		//DOF
		//uint vsCnt = uint(valhallaStations.length);
		uint vsCnt = valhallaStations.length;
		//print("vsCnt = "+vsCnt);
		file << vsCnt;
		for(uint i = 0; i < vsCnt; ++i)
			//fleets.saveAI(file, valhallaStations[i]);
			file << valhallaStations[i];
		file << vsScalingFactor;
		file << vsCurrent;
		file << vsL1;
		file << vsL2;
		file << vsL3;
		file << vsL4;
		file << vsL5;
		construction.saveConstruction(file, vsBuild);
		construction.saveConstruction(file, vsRetrofit);
	}

	void load(SaveFile& file) override {
		uint cnt = 0;
		file >> cnt;
		defaultDesigns.length = cnt;
		defaultGoals.length = cnt;
		for(uint i = 0; i < cnt; ++i) {
			file >> defaultDesigns[i];
			file >> defaultGoals[i];
		}
		
		//DOF
		file >> DesignLevel;
		uint idCnt = 0;
		file >> idCnt;
		//print("idCnt = "+idCnt);
		InventoriedDesigns.length = idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file >> InventoriedDesigns[i];
		}
		
		//DOF
		uint vsCnt = 0;
		file >> vsCnt;
		//print("vsCnt = "+vsCnt);
		//valhallaStations.length = vsCnt;
		for(uint i = 0; i < vsCnt; ++i) {
			//auto@ flAI = fleets.loadAI(file);
			//if(flAI !is null)
			//	valhallaStations.insertLast(flAI);
			//file >> valhallaStations[i];
			Object@ obj;
			file >> obj;
			if(obj !is null)  {
				valhallaStations.insertLast(obj);
			}
		}
		file >> vsScalingFactor;
		file >> vsCurrent;
		//if(!vsCurrent.valid)  @vsCurrent = null;
		file >> vsL1;
		file >> vsL2;
		file >> vsL3;
		file >> vsL4;
		file >> vsL5;
		@vsBuild = cast<BuildStation>(construction.loadConstruction(file));
		@vsRetrofit = construction.loadConstruction(file);
	}

	//Use Pre-designed ships
	bool preCompose(DesignTarget@ target) override {
		return false;
	}

	bool postCompose(DesignTarget@ target) override {
	//	auto@ d = target.designer;

	//	//Add an extra engine
	//	if(target.purpose == DP_Combat)
	//		d.composition.insertAt(0, Exhaust(tag("Engine") & tag("GivesThrust"), 0.25, 0.35));

	//	//Remove armor layers we don't need
	//	for(uint i = 0, cnt = d.composition.length; i < cnt; ++i) {
	//		if(cast<ArmorLayer>(d.composition[i]) !is null) {
	//			d.composition.removeAt(i);
	//			--i; --cnt;
	//		}
	//	}

		return false;
	}

	bool design(DesignTarget@ target, int size, const Design@& output) {
		//All designs are rescales of default designs
		const Design@ baseDesign;
		uint possible = 0;
		for(uint i = 0, cnt = defaultDesigns.length; i < cnt; ++i) {
			if(defaultGoals[i] == target.purpose) {
				possible += 1;
				if(randomd() < 1.0 / double(possible))
					@baseDesign = defaultDesigns[i];
			}
		}

		if(baseDesign is null)
			return false;

		//if(target.designer !is null) {
		//	@target.designer.baseOffDesign = baseDesign;
		//	if(target.purpose != DP_Support)
		//		@target.designer.baseOffSubsystem = sinewSubsystem;
		//	else
		//		@target.designer.baseOffSubsystem = supportSinewSubsystem;
		//	@output = target.designer.design();
		//}

		if(output is null)
			@output = scaleDesign(baseDesign, size);
		return true;
	}

	bool getRegionHasValhallaStation(Region@ reg) {
		//print("Valhalla Stations count = "+valhallaStations.length);
		if(reg is null)
			return false;
		for(uint i = 0, cnt = valhallaStations.length; i < cnt; ++i) {
			//if(valhallaStations[i].obj.region is reg)
			if(valhallaStations[i].region is reg)
				return true;
		}
		return false;
	}

	void focusTick(double time) override {
		//Monitor for design update flag
		uint DesignLevelFlag = ai.empire.getAttribute(getEmpAttribute("DesignLevelFlag"));
		if (DesignLevelFlag > DesignLevel)  {
			//print("Pre-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Pre-refresh defaultDesigns:  "+defaultDesigns.length);
			bool Scanned = false;
			for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
				const Design@ dsg = ai.empire.getDesign(i);
				
				Scanned = false;
				for(uint i = 0, cnt = InventoriedDesigns.length; i < cnt; ++i)  {
					//if(InventoriedDesigns[i].id == dsg.id)  {
					if(InventoriedDesigns[i].name == dsg.name)  {
						Scanned = true;
						break;
					}
				}
				//If the design is already inventoried then it's old and we can skip it.
				if (Scanned)
					continue;
				//If design is already a member of "AI" category, skip - it's old
				if(dsg.cls.name == "AI")
					continue;
				InventoriedDesigns.insertLast(dsg);
				
				if(dsg.newer !is null)
					continue;
				if(dsg.updated !is null)
					continue;

				uint goal = designs.classify(dsg, DP_Unknown);
				if(goal == DP_Unknown)
					continue;

				defaultDesigns.insertLast(dsg);
				defaultGoals.insertLast(goal);
			}
			DesignLevel = DesignLevelFlag;
			//print("Post-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Post-refresh defaultDesigns:  "+defaultDesigns.length);
		}
		
		//Track Valhalla Stations
		/* //Stations are not tracked as FleetAIs :(
		for(uint i = 0, cnt = fleets.fleets.length; i < cnt; ++i) {
			auto@ flAI = fleets.fleets[i];
			Ship@ ship = cast<Ship>(flAI.obj);
			//print("Fleet #"+i+" = "+ship.blueprint.design.name);
			if(ship.blueprint.design.hasTag(ST_AsgardValhallaStation))  {
				if(valhallaStations.find(flAI) == -1)  {
					//print("Queen found");
					valhallaStations.insertLast(flAI);
					//flAI.obj.autoFillSupports = false;
					//flAI.obj.allowFillFrom = true;
					//Add as a factory
					construction.registerFactory(flAI.obj);
				}
			}
		}
		*/
		//print("Station count = "+ai.empire.get_stationCount());
		auto@ data = ai.empire.getStations();
		Object@ obj;
		while(receive(data, obj)) {
			if(obj !is null)  {
				//FleetAI@ flAI = fleets.getAI(obj);
				//FleetAI@ flAI = FleetAI();
				//@flAI.obj = obj;
				//@flAI.stationed = obj.region;
				//obj.setHoldPosition(true);
				Ship@ ship = cast<Ship>(obj);
				//print("Fleet #"+i+" = "+ship.blueprint.design.name);
				if(ship.blueprint.design.hasTag(ST_AsgardValhallaStation))  {
					//if(valhallaStations.find(flAI) == -1)  {
					if(valhallaStations.find(obj) == -1)  {
						//print("Queen found");
						//valhallaStations.insertLast(flAI);
						valhallaStations.insertLast(obj);
						//flAI.obj.autoFillSupports = false;
						//flAI.obj.allowFillFrom = true;
						//Add as a factory
						construction.registerFactory(obj);
					}
				}
			}
		}

		//Cleaning up bad entries?
		for(uint i = 0, cnt = valhallaStations.length; i < cnt; ++i) {
			//Object@ obj = valhallaStations[i].obj;
			Object@ obj = valhallaStations[i];
			if(obj is null || !obj.valid || obj.owner !is ai.empire) {
				valhallaStations.removeAt(i);
				--i; --cnt;
			}
		}
		//print("Valhalla Stations count = "+valhallaStations.length);
		
		if(vsCurrent is null)  {
			//print("vsCurrent is null");
			@vsCurrent = vsL1;
		}
		//Get best Valhalla Station design
		if(vsL2 is null)  @vsL2 = ai.empire.getDesign("ASG Valhalla Station Mk II");
		if(vsL3 is null)  @vsL3 = ai.empire.getDesign("ASG Valhalla Station Mk III");
		if(vsL4 is null)  @vsL4 = ai.empire.getDesign("ASG Valhalla Station Mk IV");
		if(vsL5 is null)  @vsL5 = ai.empire.getDesign("ASG Valhalla Station Mk V");
		bool designChange = false;
		if(vsL5 !is null && vsCurrent !is vsL5)  {
			@vsCurrent = vsL5;
			designChange = true;
		}
		else if(vsL4 !is null && vsCurrent !is vsL4)  {
			@vsCurrent = vsL4;
			designChange = true;
		}
		else if(vsL3 !is null && vsCurrent !is vsL3)  {
			@vsCurrent = vsL3;
			designChange = true;
		}
		else if(vsL2 !is null && vsCurrent !is vsL2)  {
			@vsCurrent = vsL2;
			designChange = true;
		}
		//Get our budget
		double currentBudget = ai.empire.RemainingBudget + ai.empire.EstNextBudget/2;
		uint Interval = uint(floor(gameTime/900.0));
		if(Interval > vsScalingIntervalCount)  {
			vsScalingIntervalCount = Interval;			
			//print("Expected Budget = "+currentBudget);
			//Scale current Valhalla Station design
			//print("Valhalla current scalingFactor = "+vsScalingFactor);
			int designFactor = vsCurrent.size/625;
			//print("Valhalla designFactor = "+designFactor);
			int costPerFactor = vsCurrent.total(HV_BuildCost)/designFactor;
			//print("Valhalla costPerFactor = "+costPerFactor);
			int newScalingFactor = clamp((ai.empire.EstNextBudget*1.5)/costPerFactor, 1, 32);
			//print("Valhalla newScalingFactor = "+newScalingFactor);
			if(vsScalingFactor != newScalingFactor || designChange)  {
				//print("Changing current scalingFactor to "+newScalingFactor);
				if(vsScalingFactor != newScalingFactor)  vsScalingFactor = newScalingFactor;
				//print("scalingFactor is now = "+vsScalingFactor);
				auto@ newDesign = scaleDesign(vsCurrent, 625*vsScalingFactor);
				ai.empire.changeDesign(vsCurrent, newDesign, ai.empire.getDesignClass(vsCurrent.cls.name, true));
				if(vsL5 !is null)  @vsCurrent = ai.empire.getDesign("ASG Valhalla Station Mk V");
				else if(vsL4 !is null)  @vsCurrent = ai.empire.getDesign("ASG Valhalla Station Mk IV");
				else if(vsL3 !is null)  @vsCurrent = ai.empire.getDesign("ASG Valhalla Station Mk III");
				else if(vsL2 !is null)  @vsCurrent = ai.empire.getDesign("ASG Valhalla Station Mk II");
				else  @vsCurrent = ai.empire.getDesign("ASG Valhalla Station");
			}
		}
		//See if we aren't already building one and can afford to build a new Valhalla Station
		//if(vsBuild !is null && !vsBuild.completed)  print("Valhalla Station is under construction");
		//if(vsBuild !is null && !vsBuild.started)  print("Valhalla Station construction not started");
		//if(vsBuild !is null && !vsBuild.started)  {
		//	construction.cancel(vsBuild);
		//	@vsBuild = null;
		//	print("Valhalla Station construction failed");
		//}
		if(vsBuild !is null && vsBuild.completed)  @vsBuild = null;
		if(vsBuild is null && currentBudget >= vsCurrent.total(HV_BuildCost) && gameTime > 180.0)  {
			//print("Should build Valhalla Station");
			//Find a System without a Valhalla Station yet and build one		
			Region @reg;
			for(uint i = 0, cnt = military.stagingBases.length; i < cnt; ++i) {
				@reg = military.stagingBases[i].region;
				//if(getRegionHasValhallaStation(reg))  print("Staging Base #"+i+" of "+cnt+" has Valhalla Station");
				//if(!getRegionHasValhallaStation(reg))  print("Staging Base focus #"+i+" of "+cnt+" does not have Valhalla Station and chosen");
				if(reg !is null)  {
					if(!getRegionHasValhallaStation(reg))  break;
					@reg = null;
				}
			}				
			if(reg is null)  {
				for(uint i = 0, cnt = development.focuses.length; i < cnt; ++i) {
					@reg = development.focuses[i].obj.region;
					//if(getRegionHasValhallaStation(reg))  print("Development focus #"+i+" of "+cnt+" has Valhalla Station");
					//if(!getRegionHasValhallaStation(reg))  print("Development focus #"+i+" of "+cnt+" does not have Valhalla Station and chosen");
					if(reg !is null)  {
						if(!getRegionHasValhallaStation(reg))  break;
						@reg = null;
					}
				}
			}
			if(reg !is null)  {
				//print("New station location is valid for construction");				
				vec3d pos = reg.position;
				vec2d offset = random2d(reg.radius * 0.4, reg.radius * 0.8);
				pos.x += offset.x;
				pos.z += offset.y;
				@vsBuild = construction.buildStation(vsCurrent, pos, force=true); 
			}
			//else  print("No valid region found to construct Valhalla Station in");
		}
		//See if we should retrofit existing Valhalla Stations
		//if(vsRetrofit !is null && !vsRetrofit.completed)  print("Valhalla Station is retrofitting");
		if(vsRetrofit !is null && vsRetrofit.completed)  @vsRetrofit = null;
		if(vsRetrofit is null)  {
			for(uint i = 0, cnt = valhallaStations.length; i < cnt; ++i) {
				//auto@ flAI = valhallaStations[i];
				//Ship@ station = cast<Ship>(flAI.obj);
				Ship@ station = cast<Ship>(valhallaStations[i]);
				auto@ bp = station.blueprint;
				bool retrofit = false;
				//If our design is old, we should retrofit
				//if(bp.design !is vsCurrent)  retrofit = true;
				//If our current design is larger, we should retrofit
				//else if(bp.design.size < vsCurrent.size)  retrofit = true;
				//if(bp.design is vsCurrent && bp.design.size < vsCurrent.size)  retrofit = true;
				if(bp.design.size < vsCurrent.size)  retrofit = true;
				if(retrofit)  {
					//print("Found Valhalla Station suitable for retrofit");
					@factory = construction.getFactory(station.region);
					if(factory !is null)  {
						auto@ retrofit = construction.retrofit(station);
						@vsRetrofit = construction.buildNow(retrofit, factory);
					}
				}
			}
		}
	}


};

AIComponent@ createAsgard() {
	return Asgard();
}

//class Race : AIComponent {
//}
