import empire_ai.weasel.WeaselAI;
import empire_ai.weasel.race.Race;

import empire_ai.weasel.Movement;
import empire_ai.weasel.Military;
import empire_ai.weasel.Construction;
import empire_ai.weasel.Designs;
import empire_ai.weasel.Development;
import empire_ai.weasel.Systems;
import empire_ai.weasel.Budget;
//DOF
import empire_ai.weasel.Fleets;
//import empire_ai.weasel.Planets;
//from abilities import getAbilityID;
from attributes import getEmpAttribute;

from orbitals import getOrbitalModuleID;

const double MAINFRAME_MIN_DISTANCE_STAGE = 300000;
const double MAINFRAME_MIN_DISTANCE_DEVELOP = 450000;
const double MAINFRAME_MIN_TIMER = 3.0 * 60.0;
const int MAINFRAME_BUILD_MOVE_HOPS = 5;

class LinkRegion : Savable {
	Region@ region;
	Object@ obj;
	bool arrived = false;
	vec3d destination;

	void save(SaveFile& file) {
		file << region;
		file << obj;
		file << arrived;
		file << destination;
	}

	void load(SaveFile& file) {
		file >> region;
		file >> obj;
		file >> arrived;
		file >> destination;
	}
};

class Borg : Race, RaceDesigns {
	Military@ military;
	Designs@ designs;
	Construction@ construction;
	Development@ development;
	Systems@ systems;
	Budget@ budget;
	//DOF
	Fleets@ fleets;
	Movement@ movement;
	//Planets@ planets;

	array<LinkRegion@> tracked;
	array<Object@> unassigned;

	//Designs from Verdant AI
	array<const Design@> defaultDesigns;
	array<uint> defaultGoals;
	
	//DOF - Design inventory tracking
	array<const Design@> InventoriedDesigns;
	uint DesignLevel = 0;

	//BuildOrbital@ buildMainframe;
	BuildFlagship@ buildMainframe;
	//int mainframeId = -1;
	//Design@ mainframeId;
	MoveOrder@ move;
	array<FleetAI@> Queens;
	//int colonizeAbl = -1;
	//int popStatus = -1;

	double nextBuildTry = 15.0 * 60.0;

	void create() override {
		//Adjust size per money for better behavior
		ai.behavior.shipSizePerMoney = 1.0 / 11.0;
		ai.behavior.shipSizePerLabor = 1.0 / 1.6;

		@military = cast<Military>(ai.military);
		@designs = cast<Designs>(ai.designs);
		@construction = cast<Construction>(ai.construction);
		@development = cast<Development>(ai.development);
		@systems = cast<Systems>(ai.systems);
		@budget = cast<Budget>(ai.budget);
		//DOF
		@fleets = cast<Fleets>(ai.fleets);
		@movement = cast<Movement>(ai.movement);
		//@planets = cast<Planets>(ai.planets);

		//mainframeId = getOrbitalModuleID("Mainframe");
		//mainframeId = getOrbitalModuleID("BSGCylonResurrectionShip");
		//mainframeId = ai.empire.getDesign("CYL Resurrection Ship");
		//colonizeAbl = getAbilityID("CylonStationColonize");
		//popStatus = getStatusID("CylonStationPopulation");
	}

	void start() override {
		ReadLock lock(ai.empire.designMutex);
		for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
			const Design@ dsg = ai.empire.getDesign(i);
			
			//DOF - Add design to inventoried list
			InventoriedDesigns.insertLast(dsg);
			
			if(dsg.newer !is null)
				continue;
			if(dsg.updated !is null)
				continue;

			uint goal = designs.classify(dsg, DP_Unknown);
			if(goal == DP_Unknown)
				continue;

			defaultDesigns.insertLast(dsg);
			defaultGoals.insertLast(goal);
		}
	}
	
	double transferCost(double dist) {
		//return 20 + dist * 0.002;
		//DOF - Adjusting for increased galaxy size
		return 20 + dist * 0.0001;
	}

	void save(SaveFile& file) override {
		//Verdant stuff
		uint cnt = defaultDesigns.length;
		file << cnt;
		for(uint i = 0; i < cnt; ++i) {
			file << defaultDesigns[i];
			file << defaultGoals[i];
		}
		
		//DOF
		file << DesignLevel;
		uint idCnt = InventoriedDesigns.length;
		file << idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file << InventoriedDesigns[i];
		}

		construction.saveConstruction(file, buildMainframe);
		file << nextBuildTry;

		uint tCnt = tracked.length;
		file << tCnt;
		for(uint i = 0; i < tCnt; ++i)
			file << tracked[i];

		uint uCnt = unassigned.length;
		file << uCnt;
		for(uint i = 0; i < uCnt; ++i)
			file << unassigned[i];
			
		//DOF
		uint qCnt = Queens.length;
		file << qCnt;
		for(uint i = 0; i < qCnt; ++i)
			fleets.saveAI(file, Queens[i]);
	}

	void load(SaveFile& file) override {
		//Verdant stuff
		uint cnt = 0;
		file >> cnt;
		defaultDesigns.length = cnt;
		defaultGoals.length = cnt;
		for(uint i = 0; i < cnt; ++i) {
			file >> defaultDesigns[i];
			file >> defaultGoals[i];
		}
		
		//DOF
		file >> DesignLevel;
		uint idCnt = 0;
		file >> idCnt;
		InventoriedDesigns.length = idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file >> InventoriedDesigns[i];
		}

		//@buildMainframe = cast<BuildOrbital>(construction.loadConstruction(file));
		@buildMainframe = cast<BuildFlagship>(construction.loadConstruction(file));
		file >> nextBuildTry;

		uint tCnt = 0;
		file >> tCnt;
		for(uint i = 0; i < tCnt; ++i) {
			LinkRegion gt;
			file >> gt;
			tracked.insertLast(gt);
		}

		uint uCnt = 0;
		file >> uCnt;
		for(uint i = 0; i < uCnt; ++i) {
			Object@ obj;
			file >> obj;
			if(obj !is null)
				unassigned.insertLast(obj);
		}
		//DOF
		uint qCnt = 0;
		file >> qCnt;
		for(uint i = 0; i < qCnt; ++i) {
			auto@ flAI = fleets.loadAI(file);
			if(flAI !is null)
				Queens.insertLast(flAI);
		}
	}

	LinkRegion@ get(Region@ reg) {
		for(uint i = 0, cnt = tracked.length; i < cnt; ++i) {
			if(tracked[i].region is reg)
				return tracked[i];
		}
		return null;
	}

	void remove(LinkRegion@ gt) {
		if(gt.obj !is null && gt.obj.valid && gt.obj.owner is ai.empire)
			unassigned.insertLast(gt.obj);
		tracked.remove(gt);
	}

	Object@ getClosestMainframe(const vec3d& position) {
		Object@ closest;
		double minDist = INFINITY;
		for(uint i = 0, cnt = tracked.length; i < cnt; ++i) {
			Object@ obj = tracked[i].obj;
			if(obj is null)
				continue;
			if(!tracked[i].arrived)
				continue;
			double d = obj.position.distanceTo(position);
			if(d < minDist) {
				minDist = d;
				@closest = obj;
			}
		}
		return closest;
	}

	LinkRegion@ getClosestLinkRegion(const vec3d& position) {
		LinkRegion@ closest;
		double minDist = INFINITY;
		for(uint i = 0, cnt = tracked.length; i < cnt; ++i) {
			double d = tracked[i].region.position.distanceTo(position);
			if(d < minDist) {
				minDist = d;
				@closest = tracked[i];
			}
		}
		return closest;
	}

	void assignTo(LinkRegion@ gt, Object@ closest) {
		unassigned.remove(closest);
		@gt.obj = closest;
		gt.arrived = false;

		if(closest.region is gt.region)
			gt.arrived = true;
		if(!gt.arrived) {
			gt.destination = military.getStationPosition(gt.region);
			//closest.addMoveOrder(gt.destination);
			@move = cast<Movement>(ai.movement).move(closest, gt.destination);
		}
	}

	bool trackingMainframe(Object@ obj) {
		for(uint i = 0, cnt = unassigned.length; i < cnt; ++i) {
			if(unassigned[i] is obj)
				return true;
		}
		for(uint i = 0, cnt = tracked.length; i < cnt; ++i) {
			if(tracked[i].obj is obj)
				return true;
		}
		return false;
	}

	bool shouldHaveMainframe(Region@ reg, bool always = false) {
		if(military.getBase(reg) !is null)
			return true;
		if(development.isDevelopingIn(reg))
			return true;
		return false;
	}

	void focusTick(double time) override {
		//Monitor for design update flag
		uint DesignLevelFlag = ai.empire.getAttribute(getEmpAttribute("DesignLevelFlag"));
		if (DesignLevelFlag > DesignLevel)  {
			//print("Pre-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Pre-refresh defaultDesigns:  "+defaultDesigns.length);
			bool Scanned = false;
			for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
				const Design@ dsg = ai.empire.getDesign(i);
				
				Scanned = false;
				for(uint i = 0, cnt = InventoriedDesigns.length; i < cnt; ++i)  {
					//if(InventoriedDesigns[i].id == dsg.id)  {
					if(InventoriedDesigns[i].name == dsg.name)  {
						Scanned = true;
						break;
					}
				}
				//If the design is already inventoried then it's old and we can skip it.
				if (Scanned)
					continue;
				//If design is already a member of "AI" category, skip - it's old
				if(dsg.cls.name == "AI")
					continue;
				InventoriedDesigns.insertLast(dsg);
				
				if(dsg.newer !is null)
					continue;
				if(dsg.updated !is null)
					continue;

				uint goal = designs.classify(dsg, DP_Unknown);
				if(goal == DP_Unknown)
					continue;

				defaultDesigns.insertLast(dsg);
				defaultGoals.insertLast(goal);
			}
			DesignLevel = DesignLevelFlag;
			//print("Post-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Post-refresh defaultDesigns:  "+defaultDesigns.length);
		}
	
		//Manage unassigned mainframes list
		for(uint i = 0, cnt = unassigned.length; i < cnt; ++i) {
			Object@ obj = unassigned[i];
			if(obj is null || !obj.valid || obj.owner !is ai.empire) {
				unassigned.removeAt(i);
				--i; --cnt;
			}
		}

		//Detect new gates
		/*
		auto@ data = ai.empire.getOrbitals();
		Object@ obj;
		while(receive(data, obj)) {
			if(obj is null)
				continue;
			Orbital@ orb = cast<Orbital>(obj);
			if(orb is null || orb.coreModule != uint(mainframeId))
				continue;
			if(!trackingMainframe(obj))
				unassigned.insertLast(obj);
		}
		*/
		for(uint i = 0, cnt = fleets.fleets.length; i < cnt; ++i) {
			auto@ flAI = fleets.fleets[i];
			Ship@ ship = cast<Ship>(flAI.obj);
			//print("Fleet #"+i+" = "+ship.blueprint.design.name);
			if(flAI.fleetClass == FC_BorgQueen)  {
				if(Queens.find(flAI) == -1)  {
					//print("Queen found");
					Queens.insertLast(flAI);
					flAI.obj.autoFillSupports = false;
					flAI.obj.allowFillFrom = true;
				}
				if(!trackingMainframe(ship))
					unassigned.insertLast(ship);
			}
		}
		//Cleaning up bad entries?
		for(uint i = 0, cnt = Queens.length; i < cnt; ++i) {
			Object@ obj = Queens[i].obj;
			if(obj is null || !obj.valid || obj.owner !is ai.empire) {
				Queens.removeAt(i);
				--i; --cnt;
			}
		}
		//print("Total Queens = "+Queens.length);
		
				
		//If one of our Resurrection Ships is under attack, attempt to have it flee
		//??
		/*
		for(uint i = 0, cnt = Queens.length; i < cnt; ++i) {
			Object@ res = Queens[i].obj;
			if(res.inCombat) {
				Region@ staging = military.getClosestStaging();
				//Target move order to where?
			}
		}
		*/

		//Update existing gates for staging bases
		for(uint i = 0, cnt = tracked.length; i < cnt; ++i) {
			auto@ gt = tracked[i];
			bool checkAlways = false;
			if(gt.obj !is null) {
				if(!gt.obj.valid || gt.obj.owner !is ai.empire || (gt.arrived && gt.obj.region !is gt.region)) {
					@gt.obj = null;
					gt.arrived = false;
					checkAlways = true;
				}
				else if(!gt.arrived && !gt.obj.hasOrders) {
					if(gt.destination.distanceTo(gt.obj.position) < 10.0)
						gt.arrived = true;
					else
						//gt.obj.addMoveOrder(gt.destination);
						@move = cast<Movement>(ai.movement).move(gt.obj, gt.destination);
				}
			}
			if(!shouldHaveMainframe(gt.region, checkAlways)) {
				remove(tracked[i]);
				--i; --cnt;
			}
		}

		//Detect new staging bases to build mainframes at
		for(uint i = 0, cnt = military.stagingBases.length; i < cnt; ++i) {
			auto@ base = military.stagingBases[i];
			if(base.occupiedTime < MAINFRAME_MIN_TIMER)
				continue;

			if(get(base.region) is null) {
				LinkRegion@ closest = getClosestLinkRegion(base.region.position);
				if(closest !is null && closest.region.position.distanceTo(base.region.position) < MAINFRAME_MIN_DISTANCE_STAGE)
					continue;

				LinkRegion gt;
				@gt.region = base.region;
				tracked.insertLast(gt);
				break;
			}
		}

		//Detect new important planets to build mainframes at
		for(uint i = 0, cnt = development.focuses.length; i < cnt; ++i) {
			auto@ focus = development.focuses[i];
			Region@ reg = focus.obj.region;
			if(reg is null)
				continue;

			if(get(reg) is null) {
				LinkRegion@ closest = getClosestLinkRegion(reg.position);
				if(closest !is null && closest.region.position.distanceTo(reg.position) < MAINFRAME_MIN_DISTANCE_DEVELOP)
					continue;

				LinkRegion gt;
				@gt.region = reg;
				tracked.insertLast(gt);
				break;
			}
		}

		//See if we should build a new mainframe
		if(buildMainframe !is null) {
			if(buildMainframe.completed) {
				@buildMainframe = null;
				nextBuildTry = gameTime + 60.0;
			}
		}
		for(uint i = 0, cnt = tracked.length; i < cnt; ++i) {
			auto@ gt = tracked[i];
			if(gt.obj is null) {
				Object@ closest;
				double closestDist = INFINITY;
				for(uint n = 0, ncnt = unassigned.length; n < ncnt; ++n) {
					Object@ obj = unassigned[n];
					if(obj.region is gt.region) {
						@closest = obj;
						break;
					}
					if(!obj.hasMover)
						continue;
					if(buildMainframe is null && gameTime > nextBuildTry) {
						double d = obj.position.distanceTo(gt.region.position);
						if(d < closestDist) {
							closestDist = d;
							@closest = obj;
						}
					}
				}

				if(closest !is null) {
					if(log)
						ai.print("Assign mainframe to => "+gt.region.name, closest.region);
					assignTo(gt, closest);
				} else if(buildMainframe is null && gameTime > nextBuildTry) {
					if(log)
						ai.print("Build mainframe for this system", gt.region);

					bool buildLocal = true;
					auto@ factory = construction.primaryFactory;
					if(factory !is null) {
						Region@ factRegion = factory.obj.region;
						if(factRegion !is null && systems.hopDistance(gt.region, factRegion) < MAINFRAME_BUILD_MOVE_HOPS)
							buildLocal = false;
					}

					//if(buildLocal)
						//@buildMainframe = construction.buildLocalOrbital(getOrbitalModule(mainframeId));
						@buildMainframe = construction.buildFlagship(ai.empire.getDesign("BRG Queen's Diamond"), force=true);
					//else
						//@buildMainframe = construction.buildOrbital(getOrbitalModule(mainframeId), military.getStationPosition(gt.region));
						//@buildMainframe = construction.buildFlagship(ai.empire.getDesign("CYL Resurrection Ship"), military.getStationPosition(gt.region));
				}
			}
		}
	}

	//Use Pre-designed ships
	bool preCompose(DesignTarget@ target) override {
		return false;
	}

	bool postCompose(DesignTarget@ target) override {
	//	auto@ d = target.designer;

	//	//Add an extra engine
	//	if(target.purpose == DP_Combat)
	//		d.composition.insertAt(0, Exhaust(tag("Engine") & tag("GivesThrust"), 0.25, 0.35));

	//	//Remove armor layers we don't need
	//	for(uint i = 0, cnt = d.composition.length; i < cnt; ++i) {
	//		if(cast<ArmorLayer>(d.composition[i]) !is null) {
	//			d.composition.removeAt(i);
	//			--i; --cnt;
	//		}
	//	}

		return false;
	}

	bool design(DesignTarget@ target, int size, const Design@& output) {
		//All designs are rescales of default designs
		const Design@ baseDesign;
		uint possible = 0;
		for(uint i = 0, cnt = defaultDesigns.length; i < cnt; ++i) {
			if(defaultGoals[i] == target.purpose) {
				possible += 1;
				if(randomd() < 1.0 / double(possible))
					@baseDesign = defaultDesigns[i];
			}
		}

		if(baseDesign is null)
			return false;

		//if(target.designer !is null) {
		//	@target.designer.baseOffDesign = baseDesign;
		//	if(target.purpose != DP_Support)
		//		@target.designer.baseOffSubsystem = sinewSubsystem;
		//	else
		//		@target.designer.baseOffSubsystem = supportSinewSubsystem;
		//	@output = target.designer.design();
		//}

		if(output is null)
			@output = scaleDesign(baseDesign, size);
		return true;
	}

};

AIComponent@ createBorg() {
	return Borg();
}
