import empire_ai.weasel.WeaselAI;
import empire_ai.weasel.race.Race;
import empire_ai.weasel.Designs;

//DOF
from traits import getTraitID;
from attributes import getEmpAttribute;

class DOFPrebuilt : Race, RaceDesigns {
	Designs@ designs;

	//Designs from Verdant AI
	array<const Design@> defaultDesigns;
	array<uint> defaultGoals;
	
	//DOF - Design inventory tracking
	array<const Design@> InventoriedDesigns;
	uint DesignLevel = 0;

	void create() override {
		//Adjust size per money for better behavior
		if(ai.empire.hasTrait(getTraitID("Ancients")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 7.71;
			ai.behavior.shipSizePerLabor = 1.0 / 0.6;
		}
		else if(ai.empire.hasTrait(getTraitID("AncientsCannonical")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 9.05;
			ai.behavior.shipSizePerLabor = 1.0 / 0.75;
		}
		else if(ai.empire.hasTrait(getTraitID("Asgard")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 5.81;
			ai.behavior.shipSizePerLabor = 1.0 / 0.5;
		}
		else if(ai.empire.hasTrait(getTraitID("BSGColonial")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 10.00;
			ai.behavior.shipSizePerLabor = 1.0 / 1.0;
		}
		else if(ai.empire.hasTrait(getTraitID("Goauld")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 4.41;
			ai.behavior.shipSizePerLabor = 1.0 / 0.4;
		}
		else if(ai.empire.hasTrait(getTraitID("SWEmpire")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 6.19;
			ai.behavior.shipSizePerLabor = 1.0 / 0.6;
		}
		else if(ai.empire.hasTrait(getTraitID("SWRepublic")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 5.4;
			ai.behavior.shipSizePerLabor = 1.0 / 0.65;
		}
		else if(ai.empire.hasTrait(getTraitID("SWCIS")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 5.1;
			ai.behavior.shipSizePerLabor = 1.0 / 0.6;
		}
		else if(ai.empire.hasTrait(getTraitID("Tauri")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 8.41;
			ai.behavior.shipSizePerLabor = 1.0 / 0.5;
		}
		else if(ai.empire.hasTrait(getTraitID("STFederation")))  {
			ai.behavior.shipSizePerMoney = 1.0 / 15.0;
			ai.behavior.shipSizePerLabor = 1.0 / 2.9;
		}

		@designs = cast<Designs>(ai.designs);
	}

	void start() override {
		ReadLock lock(ai.empire.designMutex);
		for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
			const Design@ dsg = ai.empire.getDesign(i);
			
			//DOF - Add design to inventoried list
			InventoriedDesigns.insertLast(dsg);
			
			if(dsg.newer !is null)
				continue;
			if(dsg.updated !is null)
				continue;

			uint goal = designs.classify(dsg, DP_Unknown);
			if(goal == DP_Unknown)
				continue;

			defaultDesigns.insertLast(dsg);
			defaultGoals.insertLast(goal);
		}
	}

	void save(SaveFile& file) override {
		uint cnt = defaultDesigns.length;
		file << cnt;
		for(uint i = 0; i < cnt; ++i) {
			file << defaultDesigns[i];
			file << defaultGoals[i];
		}
		
		//DOF
		file << DesignLevel;
		uint idCnt = InventoriedDesigns.length;
		file << idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file << InventoriedDesigns[i];
		}
	}

	void load(SaveFile& file) override {
		uint cnt = 0;
		file >> cnt;
		defaultDesigns.length = cnt;
		defaultGoals.length = cnt;
		for(uint i = 0; i < cnt; ++i) {
			file >> defaultDesigns[i];
			file >> defaultGoals[i];
		}
		
		//DOF
		file >> DesignLevel;
		uint idCnt = 0;
		file >> idCnt;
		InventoriedDesigns.length = idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file >> InventoriedDesigns[i];
		}
	}

	//Use Pre-designed ships
	bool preCompose(DesignTarget@ target) override {
		return false;
	}

	bool postCompose(DesignTarget@ target) override {
	//	auto@ d = target.designer;

	//	//Add an extra engine
	//	if(target.purpose == DP_Combat)
	//		d.composition.insertAt(0, Exhaust(tag("Engine") & tag("GivesThrust"), 0.25, 0.35));

	//	//Remove armor layers we don't need
	//	for(uint i = 0, cnt = d.composition.length; i < cnt; ++i) {
	//		if(cast<ArmorLayer>(d.composition[i]) !is null) {
	//			d.composition.removeAt(i);
	//			--i; --cnt;
	//		}
	//	}

		return false;
	}

	bool design(DesignTarget@ target, int size, const Design@& output) {
		//All designs are rescales of default designs
		const Design@ baseDesign;
		uint possible = 0;
		for(uint i = 0, cnt = defaultDesigns.length; i < cnt; ++i) {
			if(defaultGoals[i] == target.purpose) {
				possible += 1;
				if(randomd() < 1.0 / double(possible))
					@baseDesign = defaultDesigns[i];
			}
		}

		if(baseDesign is null)
			return false;

		//if(target.designer !is null) {
		//	@target.designer.baseOffDesign = baseDesign;
		//	if(target.purpose != DP_Support)
		//		@target.designer.baseOffSubsystem = sinewSubsystem;
		//	else
		//		@target.designer.baseOffSubsystem = supportSinewSubsystem;
		//	@output = target.designer.design();
		//}

		if(output is null)
			@output = scaleDesign(baseDesign, size);
		return true;
	}


	void focusTick(double time) override {
		//Monitor for design update flag
		uint DesignLevelFlag = ai.empire.getAttribute(getEmpAttribute("DesignLevelFlag"));
		if (DesignLevelFlag > DesignLevel)  {
			//print("Pre-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Pre-refresh defaultDesigns:  "+defaultDesigns.length);
			bool Scanned = false;
			for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
				const Design@ dsg = ai.empire.getDesign(i);
				
				Scanned = false;
				for(uint i = 0, cnt = InventoriedDesigns.length; i < cnt; ++i)  {
					//if(InventoriedDesigns[i].id == dsg.id)  {
					if(InventoriedDesigns[i].name == dsg.name)  {
						Scanned = true;
						break;
					}
				}
				//If the design is already inventoried then it's old and we can skip it.
				if (Scanned)
					continue;
				//If design is already a member of "AI" category, skip - it's old
				if(dsg.cls.name == "AI")
					continue;
				InventoriedDesigns.insertLast(dsg);
				
				if(dsg.newer !is null)
					continue;
				if(dsg.updated !is null)
					continue;

				uint goal = designs.classify(dsg, DP_Unknown);
				if(goal == DP_Unknown)
					continue;

				defaultDesigns.insertLast(dsg);
				defaultGoals.insertLast(goal);
			}
			DesignLevel = DesignLevelFlag;
			//print("Post-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Post-refresh defaultDesigns:  "+defaultDesigns.length);
		}
	}


};

AIComponent@ createDOFPrebuilt() {
	return DOFPrebuilt();
}

//class Race : AIComponent {
//}
