import empire_ai.weasel.WeaselAI;
import empire_ai.weasel.race.Race;
import empire_ai.weasel.Designs;

import empire_ai.weasel.Colonization;
import empire_ai.weasel.Resources;
import empire_ai.weasel.Construction;
import empire_ai.weasel.Development;
import empire_ai.weasel.Fleets;
import empire_ai.weasel.Movement;
import empire_ai.weasel.Planets;
import empire_ai.weasel.Designs;

import oddity_navigation;
from abilities import getAbilityID;
from statuses import getStatusID;

//DOF
from attributes import getEmpAttribute;

class ColonizeMission : Mission {
	Planet@ target;
	MoveOrder@ move;
	double timer = 0.0;

	void save(Fleets& fleets, SaveFile& file) override {
		file << target;
		file << timer;
		fleets.movement.saveMoveOrder(file, move);
	}

	void load(Fleets& fleets, SaveFile& file) override {
		file >> target;
		file >> timer;
		@move = fleets.movement.loadMoveOrder(file);
	}

	void start(AI& ai, FleetAI& fleet) override {
		uint prior = MP_Normal;
		if(gameTime < 30.0 * 60.0)
			prior = MP_Critical;
		@move = cast<Movement>(ai.movement).move(fleet.obj, target, prior);
	}

	void tick(AI& ai, FleetAI& fleet, double time) override {
		if(move !is null) {
			if(move.failed) {
				canceled = true;
				return;
			}
			if(move.completed) {
				int ablId = cast<Goauld>(ai.race).ColonizeAbl;
				fleet.obj.activateAbilityTypeFor(ai.empire, ablId, target);

				@move = null;
				timer = gameTime + 60.0;
			}
		}
		else {
			if(target is null || !target.valid || target.quarantined
					|| (target.owner !is ai.empire && target.owner.valid)
					|| target.inCombat) {
				canceled = true;
				return;
			}

			double curPop = target.population;
			if (curPop >= 1)  {
				completed = true;
				return;
			}

			//double maxPop = max(double(target.maxPopulation), double(getPlanetLevel(target, target.primaryResourceLevel).population));
			//double curPop = target.population;
			//if(curPop >= maxPop) {
			//	completed = true;
			//	return;
			//}

			//if(gameTime >= timer) {
			//	int popStatus = cast<Wraith>(ai.race).popStatus;
			//	if(target.getStatusStackCountAny(popStatus) >= 5) {
			//		canceled = true;
			//		return;
			//	}
			//}
		}
	}
};

class Goauld : Race, RaceDesigns {
	Colonization@ colonization;
	Construction@ construction;
	Development@ development;
	Movement@ movement;
	Planets@ planets;
	Fleets@ fleets;
	Designs@ designs;

	//Designs from Verdant AI
	array<const Design@> defaultDesigns;
	array<uint> defaultGoals;
	
	//DOF - Design inventory tracking
	array<const Design@> InventoriedDesigns;
	uint DesignLevel = 0;

	DesignTarget@ mothershipDesign;
	double idleSince = 0;

	array<FleetAI@> motherships;

	int ColonizeAbl = -1;

	array<Planet@> popRequests;

	BuildFlagship@ mcBuild;
	BuildOrbital@ yardBuild;

	Factory@ factory;

	void create() override {
		//Adjust size per money for better behavior
		//ai.behavior.shipSizePerMoney = 1.0 / 4.41;
		//ai.behavior.shipSizePerLabor = 1.0 / 0.4;
		ai.behavior.shipSizePerMoney = 1.0 / 5.5;
		ai.behavior.shipSizePerLabor = 1.0 / 0.81;
		
		@colonization = cast<Colonization>(ai.colonization);
		colonization.performColonization = false;

		@development = cast<Development>(ai.development);
		development.managePlanetPressure = false;
		development.buildBuildings = false;

		@fleets = cast<Fleets>(ai.fleets);
		@construction = cast<Construction>(ai.construction);
		@planets = cast<Planets>(ai.planets);
		@designs = cast<Designs>(ai.designs);
		@movement = cast<Movement>(ai.movement);

		@ai.defs.Factory = null;
		@ai.defs.LaborStorage = null;


		//habitatAbl = getAbilityID("MothershipColonize");
		ColonizeAbl = getAbilityID("GoauldMothershipColonize");
	}

	void start() override {
		ReadLock lock(ai.empire.designMutex);
		for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
			const Design@ dsg = ai.empire.getDesign(i);
			
			//DOF - Add design to inventoried list
			InventoriedDesigns.insertLast(dsg);
			
			if(dsg.newer !is null)
				continue;
			if(dsg.updated !is null)
				continue;

			uint goal = designs.classify(dsg, DP_Unknown);
			if(goal == DP_Unknown)
				continue;

			defaultDesigns.insertLast(dsg);
			defaultGoals.insertLast(goal);
		}

		//Get the Tier 1 in our home system
		{
			ResourceSpec spec;
			spec.type = RST_Level_Specific;
			spec.level = 1;
			spec.isForImport = false;
			spec.isLevelRequirement = false;

			colonization.queueColonize(spec);
		}

		//Then find a Tier 2 to get
		{
			ResourceSpec spec;
			spec.type = RST_Level_Specific;
			spec.level = 2;
			spec.isForImport = false;
			spec.isLevelRequirement = false;

			colonization.queueColonize(spec);
		}

		//Design a mothership
		//@mothershipDesign = designs.design(DP_Mothership, 500);
		//mothershipDesign.targetMaintenance = 300;
		//mothershipDesign.targetLaborCost = 110;
		//mothershipDesign.customName = "Mothership";
		//Design a Hiveship
		//@mothershipDesign = designs.design(DP_Mothership, 1375);
		//mothershipDesign.targetMaintenance = 750;
		//mothershipDesign.targetLaborCost = 250;
		//mothershipDesign.customName = "Hiveship";
		//DOF Extras
		//MothershipScalingFactor = 1;
		//@HiveshipDesign = mothershipDesign.active;
		//Set basic Mothership design
		//@mothershipDesign = ai.empire.getDesign("GLD Ra");
	}

	uint chkInd = 0;
	void focusTick(double time) override {
		//Monitor for design update flag
		uint DesignLevelFlag = ai.empire.getAttribute(getEmpAttribute("DesignLevelFlag"));
		if (DesignLevelFlag > DesignLevel)  {
			//print("Pre-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Pre-refresh defaultDesigns:  "+defaultDesigns.length);
			bool Scanned = false;
			for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
				const Design@ dsg = ai.empire.getDesign(i);
				
				Scanned = false;
				for(uint i = 0, cnt = InventoriedDesigns.length; i < cnt; ++i)  {
					//if(InventoriedDesigns[i].id == dsg.id)  {
					if(InventoriedDesigns[i].name == dsg.name)  {
						Scanned = true;
						break;
					}
				}
				//If the design is already inventoried then it's old and we can skip it.
				if (Scanned)
					continue;
				//If design is already a member of "AI" category, skip - it's old
				if(dsg.cls.name == "AI")
					continue;
				InventoriedDesigns.insertLast(dsg);
				
				if(dsg.newer !is null)
					continue;
				if(dsg.updated !is null)
					continue;

				uint goal = designs.classify(dsg, DP_Unknown);
				if(goal == DP_Unknown)
					continue;

				defaultDesigns.insertLast(dsg);
				defaultGoals.insertLast(goal);
			}
			DesignLevel = DesignLevelFlag;
			//print("Post-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Post-refresh defaultDesigns:  "+defaultDesigns.length);
		}
	
		//Detect motherships
		for(uint i = 0, cnt = fleets.fleets.length; i < cnt; ++i) {
			auto@ flAI = fleets.fleets[i];
			Ship@ ship = cast<Ship>(flAI.obj);
			if (ship.blueprint.design.hasTag(ST_GoauldMothership))  {
				if(motherships.find(flAI) == -1) {
					//Add to list
					flAI.obj.autoFillSupports = true;
					flAI.obj.allowFillFrom = true;
					motherships.insertLast(flAI);
				}
			}
			else
				continue;
			

			//if(flAI.fleetClass != FC_Mothership)
			//if(flAI.fleetClass != FC_Mothership && flAI.fleetClass != FC_Hiveship)
			//	continue;

			//if(motherships.find(flAI) == -1) {
				//Add to our tracking list
				//flAI.obj.autoFillSupports = false;
				//flAI.obj.allowFillFrom = true;
				//motherships.insertLast(flAI);

				//Add as a factory
				//construction.registerFactory(flAI.obj);
			//}
		}

		//Cleaning up bad entries?
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			Object@ obj = motherships[i].obj;
			if(obj is null || !obj.valid || obj.owner !is ai.empire) {
				motherships.removeAt(i);
				--i; --cnt;
			}
		}

		//print(motherships.length);

		//Send motherships to do colonization
		//uint totalCount = colonization.awaitingSource.length;
		//uint motherCount = idleMothershipCount();

		for(uint i = 0, cnt = colonization.awaitingSource.length; i < cnt; ++i) {
			Planet@ dest = colonization.awaitingSource[i].target;
			if(isColonizing(dest))
				continue;

			colonizeBest(dest);
		}

		if(motherships.length == 1)
			@colonization.colonizeWeightObj = motherships[0].obj;
		else
			@colonization.colonizeWeightObj = null;

		//Try to get and keep at least 3 motherships out there relatively quickly for faster growth
		//print(gameTime);
		if (motherships.length < 2 && gameTime > 180.0)  {
			if(mcBuild !is null && mcBuild.completed)
				@mcBuild = null;
			if(mcBuild is null)  {
				//print("Requesting a new basic mothership (2) be built");
				//@mcBuild = construction.buildFlagship(mothershipDesign, force=true);
				@mcBuild = construction.buildFlagship(ai.empire.getDesign("GLD Ra"), force=true);
			}
		}
		else if (motherships.length < 3 && gameTime > 600.0)  {
			if(mcBuild !is null && mcBuild.completed)
				@mcBuild = null;
			if(mcBuild is null)  {
				//print("Requesting a new basic mothership (3) be built");
				//@mcBuild = construction.buildFlagship(mothershipDesign, force=true);
				@mcBuild = construction.buildFlagship(ai.empire.getDesign("GLD Ra"), force=true);
			}
		}
	}

	uint idleMothershipCount() {
		uint count = 0;
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			if(motherships[i].mission is null)
				count += 1;
		}
		return count;
	}

	bool isColonizing(Planet& dest) {
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			auto@ flAI = motherships[i];
			auto@ miss = cast<ColonizeMission>(flAI.mission);
			if(miss !is null && miss.target is dest)
				return true;
		}
		return false;
	}

	Planet@ colonizeBest(FleetAI& flAI) {
		Planet@ best;
		double bestDist = INFINITY;
		for(uint i = 0, cnt = popRequests.length; i < cnt; ++i) {
			Planet@ dest = popRequests[i];
			if(isColonizing(dest))
				continue;
			if(dest.inCombat)
				continue;

			double d = movement.getPathDistance(flAI.obj.position, dest.position);
			if(d < bestDist) {
				@best = dest;
				bestDist = d;
			}
		}

		if(best is null) {
			for(uint i = 0, cnt = colonization.awaitingSource.length; i < cnt; ++i) {
				Planet@ dest = colonization.awaitingSource[i].target;
				if(isColonizing(dest))
					continue;

				double d = movement.getPathDistance(flAI.obj.position, dest.position);
				if(d < bestDist) {
					@best = dest;
					bestDist = d;
				}
			}
		}

		if(best !is null) {
			ColonizeMission miss;
			@miss.target = best;

			fleets.performMission(flAI, miss);
		}
		return best;

	}

	FleetAI@ colonizeBest(Planet& dest) {
		FleetAI@ best;
		double bestDist = INFINITY;
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			auto@ flAI = motherships[i];
			if(flAI.mission !is null)
				continue;
			//if(isBuildingWithLabor(flAI))
			//	continue;

			double d = movement.getPathDistance(flAI.obj.position, dest.position);
			if(d < bestDist) {
				@best = flAI;
				bestDist = d;
			}
		}

		if(best !is null) {
			ColonizeMission miss;
			@miss.target = dest;

			fleets.performMission(best, miss);
		}
		return best;
	}

	void save(SaveFile& file) override {
		uint cnt = defaultDesigns.length;
		file << cnt;
		for(uint i = 0; i < cnt; ++i) {
			file << defaultDesigns[i];
			file << defaultGoals[i];
		}
		
		//DOF
		file << DesignLevel;
		uint idCnt = InventoriedDesigns.length;
		file << idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file << InventoriedDesigns[i];
		}

		designs.saveDesign(file, mothershipDesign);
		file << idleSince;
		construction.saveConstruction(file, mcBuild);
		construction.saveConstruction(file, yardBuild);

		uint mCnt = motherships.length;
		file << mCnt;
		for(uint i = 0; i < mCnt; ++i)
			fleets.saveAI(file, motherships[i]);
	}

	void load(SaveFile& file) override {
		uint cnt = 0;
		file >> cnt;
		defaultDesigns.length = cnt;
		defaultGoals.length = cnt;
		for(uint i = 0; i < cnt; ++i) {
			file >> defaultDesigns[i];
			file >> defaultGoals[i];
		}
		
		//DOF
		file >> DesignLevel;
		uint idCnt = 0;
		file >> idCnt;
		InventoriedDesigns.length = idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file >> InventoriedDesigns[i];
		}

		@mothershipDesign = designs.loadDesign(file);
		file >> idleSince;
		@mcBuild = cast<BuildFlagship>(construction.loadConstruction(file));
		@yardBuild = cast<BuildOrbital>(construction.loadConstruction(file));

		uint mCnt = 0;
		file >> mCnt;
		for(uint i = 0; i < mCnt; ++i) {
			auto@ flAI = fleets.loadAI(file);
			if(flAI !is null)
				motherships.insertLast(flAI);
		}
	}

	//Use Pre-designed ships
	bool preCompose(DesignTarget@ target) override {
		return false;
	}

	bool postCompose(DesignTarget@ target) override {
	//	auto@ d = target.designer;

	//	//Add an extra engine
	//	if(target.purpose == DP_Combat)
	//		d.composition.insertAt(0, Exhaust(tag("Engine") & tag("GivesThrust"), 0.25, 0.35));

	//	//Remove armor layers we don't need
	//	for(uint i = 0, cnt = d.composition.length; i < cnt; ++i) {
	//		if(cast<ArmorLayer>(d.composition[i]) !is null) {
	//			d.composition.removeAt(i);
	//			--i; --cnt;
	//		}
	//	}

		return false;
	}

	bool design(DesignTarget@ target, int size, const Design@& output) {
		//All designs are rescales of default designs
		const Design@ baseDesign;
		uint possible = 0;
		for(uint i = 0, cnt = defaultDesigns.length; i < cnt; ++i) {
			if(defaultGoals[i] == target.purpose) {
				possible += 1;
				if(randomd() < 1.0 / double(possible))
					@baseDesign = defaultDesigns[i];
			}
		}

		if(baseDesign is null)
			return false;

		//if(target.designer !is null) {
		//	@target.designer.baseOffDesign = baseDesign;
		//	if(target.purpose != DP_Support)
		//		@target.designer.baseOffSubsystem = sinewSubsystem;
		//	else
		//		@target.designer.baseOffSubsystem = supportSinewSubsystem;
		//	@output = target.designer.design();
		//}

		if(output is null)
			@output = scaleDesign(baseDesign, size);
		return true;
	}

};

AIComponent@ createGoauld() {
	return Goauld();
}

//class Race : AIComponent {
//}
