import empire_ai.weasel.WeaselAI;
import empire_ai.weasel.race.Race;
import empire_ai.weasel.Designs;

//DOF
//from traits import getTraitID;
import empire_ai.weasel.Fleets;
from attributes import getEmpAttribute;
from abilities import getAbilityID;
from statuses import getStatusID;

class Klingon : Race, RaceDesigns {
	Designs@ designs;

	//Designs from Verdant AI
	array<const Design@> defaultDesigns;
	array<uint> defaultGoals;
		
	//DOF - Design inventory tracking
	array<const Design@> InventoriedDesigns;
	uint DesignLevel = 0;
	
	Fleets@ fleets;
	
	int cloakAbl = -1;
	int cloakedStatus = -1;
	

	void create() override {
		//Adjust size per money for better behavior
		ai.behavior.shipSizePerMoney = 1.0 / 13.0;
		ai.behavior.shipSizePerLabor = 1.0 / 2.25;
		
		@designs = cast<Designs>(ai.designs);
		
		@fleets = cast<Fleets>(ai.fleets);
		
		cloakAbl = getAbilityID("CloakingGenerator");
		cloakedStatus = getStatusID("CloakedByGenerator");
	}

	void start() override {
		ReadLock lock(ai.empire.designMutex);
		for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
			const Design@ dsg = ai.empire.getDesign(i);
			
			//DOF - Add design to inventoried list
			InventoriedDesigns.insertLast(dsg);
			
			if(dsg.newer !is null)
				continue;
			if(dsg.updated !is null)
				continue;

			uint goal = designs.classify(dsg, DP_Unknown);
			if(goal == DP_Unknown)
				continue;

			defaultDesigns.insertLast(dsg);
			defaultGoals.insertLast(goal);
		}
	}

	void save(SaveFile& file) override {
		uint cnt = defaultDesigns.length;
		file << cnt;
		for(uint i = 0; i < cnt; ++i) {
			file << defaultDesigns[i];
			file << defaultGoals[i];
		}
		
		//DOF
		file << DesignLevel;
		uint idCnt = InventoriedDesigns.length;
		file << idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file << InventoriedDesigns[i];
		}
	}

	void load(SaveFile& file) override {
		uint cnt = 0;
		file >> cnt;
		defaultDesigns.length = cnt;
		defaultGoals.length = cnt;
		for(uint i = 0; i < cnt; ++i) {
			file >> defaultDesigns[i];
			file >> defaultGoals[i];
		}
		
		//DOF
		file >> DesignLevel;
		uint idCnt = 0;
		file >> idCnt;
		InventoriedDesigns.length = idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file >> InventoriedDesigns[i];
		}
	}

	//Use Pre-designed ships
	bool preCompose(DesignTarget@ target) override {
		return false;
	}

	bool postCompose(DesignTarget@ target) override {
	//	auto@ d = target.designer;

	//	//Add an extra engine
	//	if(target.purpose == DP_Combat)
	//		d.composition.insertAt(0, Exhaust(tag("Engine") & tag("GivesThrust"), 0.25, 0.35));

	//	//Remove armor layers we don't need
	//	for(uint i = 0, cnt = d.composition.length; i < cnt; ++i) {
	//		if(cast<ArmorLayer>(d.composition[i]) !is null) {
	//			d.composition.removeAt(i);
	//			--i; --cnt;
	//		}
	//	}

		return false;
	}

	bool design(DesignTarget@ target, int size, const Design@& output) {
		//All designs are rescales of default designs
		const Design@ baseDesign;
		uint possible = 0;
		for(uint i = 0, cnt = defaultDesigns.length; i < cnt; ++i) {
			if(defaultGoals[i] == target.purpose) {
				possible += 1;
				if(randomd() < 1.0 / double(possible))
					@baseDesign = defaultDesigns[i];
			}
		}

		if(baseDesign is null)
			return false;

		//if(target.designer !is null) {
		//	@target.designer.baseOffDesign = baseDesign;
		//	if(target.purpose != DP_Support)
		//		@target.designer.baseOffSubsystem = sinewSubsystem;
		//	else
		//		@target.designer.baseOffSubsystem = supportSinewSubsystem;
		//	@output = target.designer.design();
		//}

		if(output is null)
			@output = scaleDesign(baseDesign, size);
		return true;
	}


	void focusTick(double time) override {
		//Monitor for design update flag
		uint DesignLevelFlag = ai.empire.getAttribute(getEmpAttribute("DesignLevelFlag"));
		if (DesignLevelFlag > DesignLevel)  {
			//print("Pre-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Pre-refresh defaultDesigns:  "+defaultDesigns.length);
			bool Scanned = false;
			for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
				const Design@ dsg = ai.empire.getDesign(i);
				
				Scanned = false;
				for(uint i = 0, cnt = InventoriedDesigns.length; i < cnt; ++i)  {
					//if(InventoriedDesigns[i].id == dsg.id)  {
					if(InventoriedDesigns[i].name == dsg.name)  {
						Scanned = true;
						break;
					}
				}
				//If the design is already inventoried then it's old and we can skip it.
				if (Scanned)
					continue;
				//If design is already a member of "AI" category, skip - it's old
				if(dsg.cls.name == "AI")
					continue;
				InventoriedDesigns.insertLast(dsg);
				
				if(dsg.newer !is null)
					continue;
				if(dsg.updated !is null)
					continue;

				uint goal = designs.classify(dsg, DP_Unknown);
				if(goal == DP_Unknown)
					continue;

				defaultDesigns.insertLast(dsg);
				defaultGoals.insertLast(goal);
			}
			DesignLevel = DesignLevelFlag;
			//print("Post-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Post-refresh defaultDesigns:  "+defaultDesigns.length);
		}
		
		//Manage cloaks on fleet
		//Ability@ abl = getAbility(cloakAbl);
		//print("CloakAbl = "+cloakAbl);
		//cloakedStatus = getAbilityID("CloakedByGenerator");
		//print("CloakedStatus = "+cloakedStatus);
		for(uint i = 0, cnt = fleets.fleets.length; i < cnt; ++i) {
			auto@ flAI = fleets.fleets[i];
			Ship@ ship = cast<Ship>(flAI.obj);
			auto@ bp = ship.blueprint;
			//print("Fleet #"+i+" = "+ship.blueprint.design.name);
			if(bp.design.hasTag(ST_CloakGeneratorBase))  {
				//print(bp.design.name+" has Cloak Generator");
				//If not in combat and not already cloaked, activate cloak.
				//if(!flAI.obj.hasStatusEffect(cloakedStatus) && !flAI.obj.inCombat)  {					
				//if(!ship.hasStatusEffect(cloakedStatus))  {	
					//print(bp.design.name+" is not cloaked.  Activating");
					//flAI.obj.activateAbility(cloakAbl);
					//ship.activateAbilityType(cloakAbl);
					//flAI.obj.activateAbilityTypeFor(ai.empire, cloakAbl);
				//}				
				//else  print(bp.design.name+" is Cloaked");
				//Turn off cloak if engaged or attacking
				if(ship.engaged || ship.getAttackTarget() !is null || flAI.obj.inCombat)  {
					if(flAI.obj.hasStatusEffect(cloakedStatus))  {
						//We may not have to expressly turn cloak off in these circumstances, so long as we're not re-engaging it after the attack order
						flAI.obj.activateAbilityTypeFor(ai.empire, cloakAbl);
					}
				}
				else  {
					if(!flAI.obj.hasStatusEffect(cloakedStatus))  {
						flAI.obj.activateAbilityTypeFor(ai.empire, cloakAbl);
					}
				}
			}
		}
	}
};

AIComponent@ createKlingon() {
	return Klingon();
}

//class Race : AIComponent {
//}
