import empire_ai.weasel.WeaselAI;
import empire_ai.weasel.race.Race;

import empire_ai.weasel.Colonization;
import empire_ai.weasel.Resources;
import empire_ai.weasel.Construction;
import empire_ai.weasel.Development;
import empire_ai.weasel.Fleets;
import empire_ai.weasel.Movement;
import empire_ai.weasel.Planets;
import empire_ai.weasel.Designs;

import oddity_navigation;
from abilities import getAbilityID;
from statuses import getStatusID;
from attributes import getEmpAttribute;

class HabitatMission : Mission {
	Planet@ target;
	MoveOrder@ move;
	double timer = 0.0;

	void save(Fleets& fleets, SaveFile& file) override {
		file << target;
		file << timer;
		fleets.movement.saveMoveOrder(file, move);
	}

	void load(Fleets& fleets, SaveFile& file) override {
		file >> target;
		file >> timer;
		@move = fleets.movement.loadMoveOrder(file);
	}

	void start(AI& ai, FleetAI& fleet) override {
		uint prior = MP_Normal;
		if(gameTime < 30.0 * 60.0)
			prior = MP_Critical;
		@move = cast<Movement>(ai.movement).move(fleet.obj, target, prior);
	}

	void tick(AI& ai, FleetAI& fleet, double time) override {
		if(move !is null) {
			if(move.failed) {
				canceled = true;
				return;
			}
			if(move.completed) {
				int ablId = cast<Wraith>(ai.race).habitatAbl;
				fleet.obj.activateAbilityTypeFor(ai.empire, ablId, target);

				@move = null;
				timer = gameTime + 60.0;
			}
		}
		else {
			if(target is null || !target.valid || target.quarantined
					|| (target.owner !is ai.empire && target.owner.valid)
					|| target.inCombat) {
				canceled = true;
				return;
			}

			double maxPop = max(double(target.maxPopulation), double(getPlanetLevel(target, target.primaryResourceLevel).population));
			double curPop = target.population;
			if(curPop >= maxPop) {
				completed = true;
				return;
			}

			if(gameTime >= timer) {
				int popStatus = cast<Wraith>(ai.race).popStatus;
				//if(target.getStatusStackCountAny(popStatus) >= 5) {
				if(target.getStatusStackCountAny(popStatus) >= 1) {
					canceled = true;
					return;
				}
			}
		}
	}
};

class LaborMission : Mission {
	Planet@ target;
	MoveOrder@ move;
	double timer = 0.0;

	void save(Fleets& fleets, SaveFile& file) override {
		file << target;
		file << timer;
		fleets.movement.saveMoveOrder(file, move);
	}

	void load(Fleets& fleets, SaveFile& file) override {
		file >> target;
		file >> timer;
		@move = fleets.movement.loadMoveOrder(file);
	}

	void start(AI& ai, FleetAI& fleet) override {
		@move = cast<Movement>(ai.movement).move(fleet.obj, target);
	}

	void tick(AI& ai, FleetAI& fleet, double time) override {
		if(move !is null) {
			if(move.failed) {
				canceled = true;
				return;
			}
			if(move.completed) {
				@move = null;
				timer = gameTime + 10.0;
			}
		}
		else {
			if(target is null || !target.valid || target.quarantined
					|| target.owner !is ai.empire) {
				canceled = true;
				return;
			}

			if(gameTime >= timer) {
				int popStatus = cast<Wraith>(ai.race).popStatus;
				timer = gameTime + 10.0;
				//if(target.getStatusStackCountAny(popStatus) >= 10) {
				if(target.getStatusStackCountAny(popStatus) >= 1) {
					completed = true;
					return;
				}
			}
		}
	}
};

class Wraith : Race, RaceDesigns {
	Colonization@ colonization;
	Construction@ construction;
	Development@ development;
	Movement@ movement;
	Planets@ planets;
	Fleets@ fleets;
	Designs@ designs;

	//Designs from Verdant AI
	array<const Design@> defaultDesigns;
	array<uint> defaultGoals;
	
	//DOF - Design inventory tracking
	array<const Design@> InventoriedDesigns;
	uint DesignLevel = 0;

	const SubsystemDef@ sinewSubsystem;
	const SubsystemDef@ supportSinewSubsystem;

	DesignTarget@ mothershipDesign;
	double idleSince = 0;

	array<FleetAI@> motherships;

	int habitatAbl = -1;
	int popStatus = -1;

	array<Planet@> popRequests;
	array<Planet@> laborPlanets;

	BuildFlagship@ mcBuild;
	BuildOrbital@ yardBuild;

	//DOF Extras
	int MothershipScalingFactor = 0;
	int CurrentBudget = 0;
	const Design@ HiveshipDesign;
	AllocateConstruction@ HiveshipRetrofit;
	Factory@ factory;

	void save(SaveFile& file) override {
		//Verdant stuff
		uint cnt = defaultDesigns.length;
		file << cnt;
		for(uint i = 0; i < cnt; ++i) {
			file << defaultDesigns[i];
			file << defaultGoals[i];
		}
		
		//DOF
		file << DesignLevel;
		uint idCnt = InventoriedDesigns.length;
		file << idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file << InventoriedDesigns[i];
		}

		designs.saveDesign(file, mothershipDesign);
		file << idleSince;
		construction.saveConstruction(file, mcBuild);
		construction.saveConstruction(file, yardBuild);

		uint mCnt = motherships.length;
		mCnt = motherships.length;
		file << mCnt;
		for(uint i = 0; i < mCnt; ++i)
			fleets.saveAI(file, motherships[i]);
		//DOF
		file << MothershipScalingFactor;
		construction.saveConstruction(file, HiveshipRetrofit);
	}

	void load(SaveFile& file) override {
		//Verdant stuff
		uint cnt = 0;
		file >> cnt;
		defaultDesigns.length = cnt;
		defaultGoals.length = cnt;
		for(uint i = 0; i < cnt; ++i) {
			file >> defaultDesigns[i];
			file >> defaultGoals[i];
		}
		
		//DOF
		file >> DesignLevel;
		uint idCnt = 0;
		file >> idCnt;
		InventoriedDesigns.length = idCnt;
		for(uint i = 0; i < idCnt; ++i) {
			file >> InventoriedDesigns[i];
		}

		@mothershipDesign = designs.loadDesign(file);
		file >> idleSince;
		@mcBuild = cast<BuildFlagship>(construction.loadConstruction(file));
		@yardBuild = cast<BuildOrbital>(construction.loadConstruction(file));

		uint mCnt = 0;
		mCnt = 0;
		file >> mCnt;
		for(uint i = 0; i < mCnt; ++i) {
			auto@ flAI = fleets.loadAI(file);
			if(flAI !is null)
				motherships.insertLast(flAI);
		}
		//DOF
		file >> MothershipScalingFactor;
		@HiveshipRetrofit = construction.loadConstruction(file);
	}

	void create() override {
		//Adjust size per money for better behavior
		ai.behavior.shipSizePerMoney = 1.0 / 4.81;
		ai.behavior.shipSizePerLabor = 1.0 / 0.45;

		@colonization = cast<Colonization>(ai.colonization);
		colonization.performColonization = false;

		@development = cast<Development>(ai.development);
		development.managePlanetPressure = false;
		development.buildBuildings = false;

		@fleets = cast<Fleets>(ai.fleets);
		@construction = cast<Construction>(ai.construction);
		@planets = cast<Planets>(ai.planets);
		@designs = cast<Designs>(ai.designs);
		@movement = cast<Movement>(ai.movement);

		@ai.defs.Factory = null;
		@ai.defs.LaborStorage = null;

		@sinewSubsystem = getSubsystemDef("WraithSinew");
		@supportSinewSubsystem = getSubsystemDef("WraithSinewSupport");

		//habitatAbl = getAbilityID("MothershipColonize");
		//popStatus = getAbilityID("MothershipPopulation");
		habitatAbl = getAbilityID("WraithHiveshipColonize");
		//popStatus = getAbilityID("WraithHiveshipPopulation");
		popStatus = getStatusID("WraithHiveshipPopulation");
	}

	void start() override {
		//Verdant stuff
		ReadLock lock(ai.empire.designMutex);
		for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
			const Design@ dsg = ai.empire.getDesign(i);
			
			//DOF - Add design to inventoried list
			InventoriedDesigns.insertLast(dsg);
			
			if(dsg.newer !is null)
				continue;
			if(dsg.updated !is null)
				continue;

			uint goal = designs.classify(dsg, DP_Unknown);
			if(goal == DP_Unknown)
				continue;

			defaultDesigns.insertLast(dsg);
			defaultGoals.insertLast(goal);
		}

		//Get the Tier 1 in our home system
		{
			ResourceSpec spec;
			spec.type = RST_Level_Specific;
			spec.level = 1;
			spec.isForImport = false;
			spec.isLevelRequirement = false;

			colonization.queueColonize(spec);
		}

		//Then find a Tier 2 to get
		{
			ResourceSpec spec;
			spec.type = RST_Level_Specific;
			spec.level = 2;
			spec.isForImport = false;
			spec.isLevelRequirement = false;

			colonization.queueColonize(spec);
		}

		//Design a mothership
		//@mothershipDesign = designs.design(DP_Mothership, 500);
		//mothershipDesign.targetMaintenance = 300;
		//mothershipDesign.targetLaborCost = 110;
		//mothershipDesign.customName = "Mothership";
		//Design a Hiveship
		@mothershipDesign = designs.design(DP_Mothership, 1375);
		mothershipDesign.targetMaintenance = 625;
		mothershipDesign.targetLaborCost = 145;
		mothershipDesign.customName = "Hiveship";
		//DOF Extras
		MothershipScalingFactor = 1;
		@HiveshipDesign = mothershipDesign.active;
	}

	bool requiresPopulation(Planet& target) {
		double maxPop = max(double(target.maxPopulation), double(getPlanetLevel(target, target.primaryResourceLevel).population));
		double curPop = target.population;
		return curPop < maxPop;
	}

	uint chkInd = 0;
	void focusTick(double time) override {
		//Monitor for design update flag
		uint DesignLevelFlag = ai.empire.getAttribute(getEmpAttribute("DesignLevelFlag"));
		if (DesignLevelFlag > DesignLevel)  {
			//print("Pre-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Pre-refresh defaultDesigns:  "+defaultDesigns.length);
			bool Scanned = false;
			for(uint i = 0, cnt = ai.empire.designCount; i < cnt; ++i) {
				const Design@ dsg = ai.empire.getDesign(i);
				
				Scanned = false;
				for(uint i = 0, cnt = InventoriedDesigns.length; i < cnt; ++i)  {
					//if(InventoriedDesigns[i].id == dsg.id)  {
					if(InventoriedDesigns[i].name == dsg.name)  {
						Scanned = true;
						break;
					}
				}
				//If the design is already inventoried then it's old and we can skip it.
				if (Scanned)
					continue;
				//If design is already a member of "AI" category, skip - it's old
				if(dsg.cls.name == "AI")
					continue;
				InventoriedDesigns.insertLast(dsg);
				
				if(dsg.newer !is null)
					continue;
				if(dsg.updated !is null)
					continue;

				uint goal = designs.classify(dsg, DP_Unknown);
				if(goal == DP_Unknown)
					continue;

				defaultDesigns.insertLast(dsg);
				defaultGoals.insertLast(goal);
			}
			DesignLevel = DesignLevelFlag;
			//print("Post-refresh InventoriedDesigns:  "+InventoriedDesigns.length);
			//print("Post-refresh defaultDesigns:  "+defaultDesigns.length);
		}
	
		//Detect motherships
		for(uint i = 0, cnt = fleets.fleets.length; i < cnt; ++i) {
			auto@ flAI = fleets.fleets[i];
			//if(flAI.fleetClass != FC_Mothership)
			if(flAI.fleetClass != FC_Mothership && flAI.fleetClass != FC_Hiveship)
				continue;

			if(motherships.find(flAI) == -1) {
				//Add to our tracking list
				//flAI.obj.autoFillSupports = false;
				//DOF - changing this so Hiveships auto-fill as they should
				flAI.obj.autoFillSupports = true;
				flAI.obj.allowFillFrom = true;
				motherships.insertLast(flAI);

				//Add as a factory
				construction.registerFactory(flAI.obj);
			}
		}

		//Cleaning up bad entries?
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			Object@ obj = motherships[i].obj;
			if(obj is null || !obj.valid || obj.owner !is ai.empire) {
				motherships.removeAt(i);
				--i; --cnt;
			}
		}

		//Re-classify some motherships as Hiveships
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			auto@ flAI = fleets.getAI(motherships[i].obj);
			//if(i <= max(motherships.length/2, 2)) {
			if(i <= min(motherships.length/2, 2)) {
				if (flAI.fleetClass != FC_Mothership)  {
					flAI.fleetClass = FC_Mothership;
					//print("Hiveship # "+i+" re-classified as FC_Mothership");
				}
			}
			else  {
				if (flAI.fleetClass != FC_Hiveship)  {
					flAI.fleetClass = FC_Hiveship;
					//print("Hiveship # "+i+" re-classified as FC_Hiveship");
				}
			}
		}

		//Detect planets that require more population
		for(uint i = 0, cnt = popRequests.length; i < cnt; ++i) {
			auto@ obj = popRequests[i];
			if(obj is null || !obj.valid || obj.owner !is ai.empire) {
				popRequests.removeAt(i);
				--i; --cnt;
				continue;
			}
			if(!requiresPopulation(obj)) {
				popRequests.removeAt(i);
				--i; --cnt;
				continue;
			}
		}

		for(uint i = 0, cnt = laborPlanets.length; i < cnt; ++i) {
			auto@ obj = laborPlanets[i];
			if(obj is null || !obj.valid || obj.owner !is ai.empire) {
				laborPlanets.removeAt(i);
				--i; --cnt;
				continue;
			}
			if(obj.laborIncome < 3.0/60.0) {
				laborPlanets.removeAt(i);
				--i; --cnt;
				continue;
			}
		}

		uint plCnt = planets.planets.length;
		for(uint n = 0, cnt = min(15, plCnt); n < cnt; ++n) {
			chkInd = (chkInd+1) % plCnt;
			auto@ plAI = planets.planets[chkInd];

			//Find planets that need population
			if(requiresPopulation(plAI.obj)) {
				if(popRequests.find(plAI.obj) == -1)
					popRequests.insertLast(plAI.obj);
			}

			//Find planets that have labor
			if(plAI.obj.laborIncome >= 3.0/60.0) {
				if(laborPlanets.find(plAI.obj) == -1)
					laborPlanets.insertLast(plAI.obj);
			}
		}

		//Send motherships to do colonization
		uint totalCount = popRequests.length + colonization.awaitingSource.length;
		uint motherCount = idleMothershipCount();

		/*if(motherCount > totalCount) {*/

			for(uint i = 0, cnt = popRequests.length; i < cnt; ++i) {
				Planet@ dest = popRequests[i];
				if(isColonizing(dest))
					continue;
				if(dest.inCombat)
					continue;

				colonizeBest(dest);
			}

			for(uint i = 0, cnt = colonization.awaitingSource.length; i < cnt; ++i) {
				Planet@ dest = colonization.awaitingSource[i].target;
				if(isColonizing(dest))
					continue;

				colonizeBest(dest);
			}

		/*}*/
		/*else {*/
		/*	for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {*/
		/*		auto@ flAI = motherships[i];*/
		/*		if(flAI.mission !is null)*/
		/*			continue;*/
		/*		if(isBuildingWithLabor(flAI))*/
		/*			continue;*/

		/*		colonizeBest(flAI);*/
		/*	}*/
		/*}*/

		if(totalCount != 0)
			idleSince = gameTime;

/*  Need to redo this
		//See if we should build new motherships
		uint haveMC = motherships.length;
		uint wantMC = 1;
		if(gameTime > 20.0 * 60.0)
			wantMC += 1;
		wantMC = max(wantMC, min(laborPlanets.length, uint(gameTime/(30.0*60.0))));

		if(mcBuild !is null && mcBuild.completed)
			@mcBuild = null;
		if(wantMC > haveMC && mcBuild is null)
			@mcBuild = construction.buildFlagship(mothershipDesign, force=true);

		if(yardBuild is null && haveMC > 0 && gameTime > 60 && gameTime < 180 && ai.defs.Shipyard !is null) {
			Region@ reg = motherships[0].obj.region;
			if(reg !is null) {
				vec3d pos = reg.position;
				vec2d offset = random2d(reg.radius * 0.4, reg.radius * 0.8);
				pos.x += offset.x;
				pos.z += offset.y;

				@yardBuild = construction.buildOrbital(ai.defs.Shipyard, pos);
			}
		}
*/
		//Revised See if we should build new motherships
		uint haveMC = motherships.length;
		//uint wantMC = 1;
		uint wantMC = haveMC;
		//Empire@ emp = motherships[0].obj.owner;
		//uint CurrentMotherships = emp.getAttribute(TotalHiveships.integer);
		//double TotalPlanetPopulation = motherships[0].obj.owner.TotalPopulation;
		double TotalPlanetPopulation = ai.empire.TotalPopulation;
		//print(TotalPlanetPopulation);
		if(uint(TotalPlanetPopulation) > (haveMC * 24))
			wantMC += 1;
			//wantMC += uint(TotalPlanetPopulation)/(haveMC * 16);
		//wantMC = max(wantMC, min(laborPlanets.length, uint(gameTime/(30.0*60.0))));

		if(mcBuild !is null && mcBuild.completed)
			@mcBuild = null;
		if(wantMC > haveMC && mcBuild is null)  {
			//print("Requesting a new Hiveship be built");
			@mcBuild = construction.buildFlagship(mothershipDesign, force=true);
		}

		if(yardBuild is null && haveMC > 0 && gameTime > 60 && gameTime < 180 && ai.defs.Shipyard !is null) {
		//if(yardBuild is null && haveMC > 0 && ai.defs.Shipyard !is null) {
			Region@ reg = motherships[0].obj.region;
			if(reg !is null) {
				vec3d pos = reg.position;
				vec2d offset = random2d(reg.radius * 0.4, reg.radius * 0.8);
				pos.x += offset.x;
				pos.z += offset.y;

				@yardBuild = construction.buildOrbital(ai.defs.Shipyard, pos);
			}
		}

		/*
		//Try to isolate error causes
		if(HiveshipDesign !is null && HiveshipDesign.hasFatalErrors())  {
			for(uint i = 0, cnt = HiveshipDesign.errorCount; i < cnt; ++i)
				print("   "+HiveshipDesign.errors[i].text);
		}
		//Scaling up mothership design
		@HiveshipDesign = mothershipDesign.active;
		CurrentBudget = ai.empire.EstNextBudget/2 + ai.empire.RemainingBudget;  //Or ai.empire.RemainingBudget?
		//print("Budget = "+CurrentBudget);
		//if (HiveshipDesign is null)  print("HiveshipDesign is null");
		if (HiveshipDesign !is null && CurrentBudget >= HiveshipDesign.total(HV_BuildCost) * 2.25)  {
		//We have the budget, so double design size
			//print("Budget = "+CurrentBudget);
			//print("Scale-up:  Hiveship design cost = "+HiveshipDesign.total(HV_BuildCost));
			MothershipScalingFactor = MothershipScalingFactor * 2;
			mothershipDesign = designs.scale(HiveshipDesign, 1375 * MothershipScalingFactor);
			mothershipDesign.targetMaintenance = 750 * MothershipScalingFactor;
			mothershipDesign.targetLaborCost = 250 * MothershipScalingFactor;
			mothershipDesign.customName = "Hiveship";
			@HiveshipDesign = mothershipDesign.active;
		}
		else if (HiveshipDesign !is null && MothershipScalingFactor > 1 && CurrentBudget < HiveshipDesign.total(HV_BuildCost) * 0.666667 && ai.empire.RemainingBudget > ai.empire.EstNextBudget)  {
		//We can't afford to build the current design so scale down if already scaled up
			//print("Budget = "+CurrentBudget);
			//print("Scale-down:  Hiveship design cost = "+HiveshipDesign.total(HV_BuildCost));
			//MothershipScalingFactor /= 2;
			MothershipScalingFactor = MothershipScalingFactor / 2;
			mothershipDesign = designs.scale(HiveshipDesign, 1375 * MothershipScalingFactor);
			mothershipDesign.targetMaintenance = 750 * MothershipScalingFactor;
			mothershipDesign.targetLaborCost = 250 * MothershipScalingFactor;		
			mothershipDesign.customName = "Hiveship";
			@HiveshipDesign = mothershipDesign.active;
		}
		*/
		
		//Try to isolate error causes
		//@HiveshipDesign = mothershipDesign.active;
		/*
		if(mothershipDesign.active is null && gameTime > 300.0)  {
			//print("mothershipDesign.active current design = "+mothershipDesign.active.name);
			print("Warning:  mothershipDesign.active = null.  Danger Wraith Robinson!");
			//@HiveshipDesign = ai.empire.getDesign("WRA Hiveship");
			//print("Rebuilding with base of:  "+HiveshipDesign.name);
			//print("WRA Hiveship size = "+HiveshipDesign.size);
			print("Attempting to generate new mothershipDesign");
			@mothershipDesign = designs.design(DP_Mothership, 1375);
			mothershipDesign.targetMaintenance = 625;
			mothershipDesign.targetLaborCost = 145;
			mothershipDesign.customName = "Hiveship";
			//ai.empire.changeDesign(HiveshipDesign, mothershipDesign.active, ai.empire.getDesignClass("AI", true));
			//@mothershipDesign.active = HiveshipDesign;
			//@mothershipDesign.active = scaleDesign(HiveshipDesign, 1375);
			//print("mothershipDesign.active set to:  "+HiveshipDesign.name);
			
		}
		else if(mothershipDesign.active !is null && gameTime > 300.0)  {
			print("mothershipDesign.active design exists");
			print("mothershipDesign.active name = "+mothershipDesign.active.name);
			print("mothershipDesign.active size = "+mothershipDesign.active.size);
		}		
		@HiveshipDesign = mothershipDesign.active;
		if(HiveshipDesign !is null && HiveshipDesign.hasFatalErrors())  {
			//print("HiveshipDesign:  "+HiveshipDesign.name);
			for(uint i = 0, cnt = HiveshipDesign.errorCount; i < cnt; ++i)
				print("   "+HiveshipDesign.errors[i].text);
			print("Design Size = "+HiveshipDesign.size);
			//@HiveshipDesign = ai.empire.getDesign("WRA Hiveship");
			//print("Checking base design:  "+HiveshipDesign.name);
			//for(uint i = 0, cnt = HiveshipDesign.errorCount; i < cnt; ++i)
				//print("   "+HiveshipDesign.errors[i].text);
			//print("Design Size = "+HiveshipDesign.size);
			//@HiveshipDesign = scaleDesign(HiveshipDesign, 1375);
			//print("Design Size = "+HiveshipDesign.size);
			//Attemp to fix
			//print("Attempting to rebuild bad Hiveship design");
			//@HiveshipDesign = ai.empire.getDesign("WRA Hiveship");
			//int MothershipSize = HiveshipDesign.size * MothershipScalingFactor;
			//DesignTarget target(DP_Mothership, MothershipSize);
			//@target.active = scaleDesign(HiveshipDesign, MothershipSize);
			//@HiveshipDesign = mothershipDesign.active;
			//ai.empire.changeDesign(HiveshipDesign, target.active, ai.empire.getDesignClass("AI", true));
		}
		//print("Active MothershipDesign:  "+mothershipDesign.active.name);
		*/
		
		//Scaling up mothership design
		//print("MothershipScalingFactor = "+MothershipScalingFactor);
		//bool inProgress = designs.isDesigning(mothershipDesign);
		//if(inProgress)  print("isDesigning mothershipDesign = true");
		CurrentBudget = ai.empire.EstNextBudget/2 + ai.empire.RemainingBudget;  //Or ai.empire.RemainingBudget?
		//print("Budget = "+CurrentBudget);
		@HiveshipDesign = mothershipDesign.active;
		//if (HiveshipDesign is null)  print("HiveshipDesign is null");
		//if (HiveshipDesign !is null && CurrentBudget >= HiveshipDesign.total(HV_BuildCost) * 2.25 && !inProgress)  {
		if (HiveshipDesign !is null && CurrentBudget >= HiveshipDesign.total(HV_BuildCost) * 2.25)  {
		//We have the budget, so double design size
			//print("Budget = "+CurrentBudget);
			//print("Scale-up:  Hiveship design cost = "+HiveshipDesign.total(HV_BuildCost));
			MothershipScalingFactor = max(MothershipScalingFactor * 2, 1);
			//print("MothershipScalingFactor = "+MothershipScalingFactor);
			mothershipDesign = designs.scale(HiveshipDesign, 1375 * MothershipScalingFactor);
			mothershipDesign.targetMaintenance = 750 * MothershipScalingFactor;
			mothershipDesign.targetLaborCost = 250 * MothershipScalingFactor;
			mothershipDesign.customName = "Hiveship";
			@HiveshipDesign = mothershipDesign.active;			
		}
		//else if (HiveshipDesign !is null && MothershipScalingFactor > 1 && CurrentBudget < HiveshipDesign.total(HV_BuildCost) * 0.666667 && ai.empire.RemainingBudget > ai.empire.EstNextBudget && !inProgress)  {
		else if (HiveshipDesign !is null && MothershipScalingFactor > 1 && CurrentBudget < HiveshipDesign.total(HV_BuildCost) * 0.666667 && ai.empire.RemainingBudget > ai.empire.EstNextBudget)  {
		//We can't afford to build the current design so scale down if already scaled up
			//print("Budget = "+CurrentBudget);
			//print("Scale-down:  Hiveship design cost = "+HiveshipDesign.total(HV_BuildCost));
			//MothershipScalingFactor /= 2;
			MothershipScalingFactor = max(MothershipScalingFactor / 2, 1);
			mothershipDesign = designs.scale(HiveshipDesign, 1375 * MothershipScalingFactor);
			mothershipDesign.targetMaintenance = 750 * MothershipScalingFactor;
			mothershipDesign.targetLaborCost = 250 * MothershipScalingFactor;		
			mothershipDesign.customName = "Hiveship";
			@HiveshipDesign = mothershipDesign.active;
		}

		//Retrofit motherships
		if(HiveshipRetrofit !is null && HiveshipRetrofit.completed)  {
			@HiveshipRetrofit = null;
			//print("Reset HiveshipRetrofit construction to null");
		}
		//if (HiveshipDesign !is null && mcBuild is null)  {
		//if (HiveshipDesign !is null)  {
		if (HiveshipDesign !is null  && HiveshipRetrofit is null)  {
			//Ship@ Hiveship = cast<Ship>(motherships[0].obj);
			for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
				Ship@ ship = cast<Ship>(motherships[i].obj);
				if(ship.blueprint.design.name == "Hiveship" && ship.blueprint.design.size < HiveshipDesign.size)  {
					//retrofit
					//print("Requesting retrofitting of hiveship# "+i);
					auto@ retrofit = construction.retrofit(ship);
					//for now assign the retrofit to ourself
					@factory = construction.getFactory(ship.region);
					//if (factory is null)  print("Factory for retrofit is null");
					@HiveshipRetrofit = construction.buildNow(retrofit, factory);
					//if (HiveshipRetrofit is null)  print("HiveshipRetrofit is null");
					break;
				}
			}
		}

		//Handle excessive motherships
		if(haveMC > 1 && uint(TotalPlanetPopulation) < ((haveMC-1) * 24 - 12))  {
		//We have more motherships than our population can support so our economy is sub-optimal.  Scuttle one
			//print("Too many Hiveships for population:  starting scuttle procedure.");
			int smallestSize = 0;
			Ship@ smallestShip = cast<Ship>(motherships[0].obj);
			for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
				Ship@ ship = cast<Ship>(motherships[i].obj);
				if(ship.blueprint.design.size < smallestShip.blueprint.design.size) smallestSize = i;
				}
			smallestShip.destroy();
			motherships.removeAt(smallestSize);
		}

		if(motherships.length == 1)
			@colonization.colonizeWeightObj = motherships[0].obj;
		else
			@colonization.colonizeWeightObj = null;

		//Idle motherships should be sent to go collect labor from labor planets
		if(laborPlanets.length != 0) {
			for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
				auto@ flAI = motherships[i];
				if(flAI.mission !is null)
					continue;
				if(isAtLaborPlanet(flAI))
					continue;
				//if(i == 0 && idleSince < gameTime-60.0)
				//DOF - Never allow our only mothership to get locked into a labor mission.
				if(i == 0 && motherships.length == 1)
					continue;

				double bestDist = INFINITY;
				Planet@ best;
				for(uint n = 0, ncnt = laborPlanets.length; n < ncnt; ++n) {
					Planet@ check = laborPlanets[n];
					if(hasMothershipAt(check))
						continue;

					double d = movement.getPathDistance(flAI.obj.position, check.position);
					if(d < bestDist) {
						@best = check;
						bestDist = d;
					}
				}

				if(best !is null) {
					LaborMission miss;
					@miss.target = best;

					fleets.performMission(flAI, miss);
				}
			}
		}
	}

	bool isAtLaborPlanet(FleetAI& flAI) {
		auto@ miss = cast<LaborMission>(flAI);
		if(miss !is null)
			return true;

		for(uint i = 0, cnt = laborPlanets.length; i < cnt; ++i) {
			if(flAI.obj.isLockedOrbit(laborPlanets[i]))
				return true;
		}
		return false;
	}

	bool isBuildingWithLabor(FleetAI& flAI) {
		auto@ f = construction.get(flAI.obj);
		if(f !is null && f.active !is null)
			return false;
		if(isAtLaborPlanet(flAI))
			return true;
		return false;
	}

	bool hasMothershipAt(Planet& pl) {
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			auto@ flAI = motherships[i];

			auto@ miss = cast<LaborMission>(flAI);
			if(miss !is null && miss.target is pl)
				return true;

			if(flAI.obj.isLockedOrbit(pl))
				return true;
		}
		return false;
	}

	uint idleMothershipCount() {
		uint count = 0;
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			if(motherships[i].mission is null)
				count += 1;
		}
		return count;
	}

	bool isColonizing(Planet& dest) {
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			auto@ flAI = motherships[i];
			auto@ miss = cast<HabitatMission>(flAI.mission);
			if(miss !is null && miss.target is dest)
				return true;
		}
		return false;
	}

	Planet@ colonizeBest(FleetAI& flAI) {
		Planet@ best;
		double bestDist = INFINITY;
		for(uint i = 0, cnt = popRequests.length; i < cnt; ++i) {
			Planet@ dest = popRequests[i];
			if(isColonizing(dest))
				continue;
			if(dest.inCombat)
				continue;

			double d = movement.getPathDistance(flAI.obj.position, dest.position);
			if(d < bestDist) {
				@best = dest;
				bestDist = d;
			}
		}

		if(best is null) {
			for(uint i = 0, cnt = colonization.awaitingSource.length; i < cnt; ++i) {
				Planet@ dest = colonization.awaitingSource[i].target;
				if(isColonizing(dest))
					continue;

				double d = movement.getPathDistance(flAI.obj.position, dest.position);
				if(d < bestDist) {
					@best = dest;
					bestDist = d;
				}
			}
		}

		if(best !is null) {
			HabitatMission miss;
			@miss.target = best;

			fleets.performMission(flAI, miss);
		}
		return best;

	}

	FleetAI@ colonizeBest(Planet& dest) {
		FleetAI@ best;
		double bestDist = INFINITY;
		for(uint i = 0, cnt = motherships.length; i < cnt; ++i) {
			auto@ flAI = motherships[i];
			if(flAI.mission !is null)
				continue;
			if(isBuildingWithLabor(flAI))
				continue;

			double d = movement.getPathDistance(flAI.obj.position, dest.position);
			if(d < bestDist) {
				@best = flAI;
				bestDist = d;
			}
		}

		if(best !is null) {
			HabitatMission miss;
			@miss.target = dest;

			fleets.performMission(best, miss);
		}
		return best;
	}

	//DOF - Attempting to override generic weighting
	double getGenericUsefulness(const ResourceType@ type) {
		//We should just colonize everything since as the Wraith we don't care and it all benefits us.
		return 100.0;
	}

	//Use Pre-designed ships
	bool preCompose(DesignTarget@ target) override {
		return false;
	}

	bool postCompose(DesignTarget@ target) override {
	//	auto@ d = target.designer;

	//	//Add an extra engine
	//	if(target.purpose == DP_Combat)
	//		d.composition.insertAt(0, Exhaust(tag("Engine") & tag("GivesThrust"), 0.25, 0.35));

	//	//Remove armor layers we don't need
	//	for(uint i = 0, cnt = d.composition.length; i < cnt; ++i) {
	//		if(cast<ArmorLayer>(d.composition[i]) !is null) {
	//			d.composition.removeAt(i);
	//			--i; --cnt;
	//		}
	//	}

		return false;
	}

	bool design(DesignTarget@ target, int size, const Design@& output) {
		//All designs are rescales of default designs
		const Design@ baseDesign;
		uint possible = 0;
		for(uint i = 0, cnt = defaultDesigns.length; i < cnt; ++i) {
			if(defaultGoals[i] == target.purpose) {
				possible += 1;
				if(randomd() < 1.0 / double(possible))
					@baseDesign = defaultDesigns[i];
			}
		}

		if(baseDesign is null)
			return false;

		//if(target.designer !is null) {
		//	@target.designer.baseOffDesign = baseDesign;
		//	if(target.purpose != DP_Support)
		//		@target.designer.baseOffSubsystem = sinewSubsystem;
		//	else
		//		@target.designer.baseOffSubsystem = supportSinewSubsystem;
		//	@output = target.designer.design();
		//}

		if(output is null)
			@output = scaleDesign(baseDesign, size);
		return true;
	}
	
	//DOF
	bool designHiveship(DesignTarget@ target, string TargetName, const Design@& output) {
		//All designs are rescales of default designs
		const Design@ baseDesign;
		uint possible = 0;
		for(uint i = 0, cnt = defaultDesigns.length; i < cnt; ++i) {
			if(defaultGoals[i] == target.purpose) {
				//possible += 1;
				//if(randomd() < 1.0 / double(possible))
				//	@baseDesign = defaultDesigns[i];
				if (defaultDesigns[i].name == TargetName)  {
					@baseDesign = defaultDesigns[i];
					break;
				}
			}
		}

		if(baseDesign is null)
			return false;

		//if(target.designer !is null) {
		//	@target.designer.baseOffDesign = baseDesign;
		//	if(target.purpose != DP_Support)
		//		@target.designer.baseOffSubsystem = sinewSubsystem;
		//	else
		//		@target.designer.baseOffSubsystem = supportSinewSubsystem;
		//	@output = target.designer.design();
		//}

		if(output is null)
			@output = scaleDesign(baseDesign, baseDesign.size);
		return true;
	}
};

AIComponent@ createWraith() {
	return Wraith();
}
